"""
DBOD InnoDB Cluster
"""

from time import time
from dbod_plugin import DbodCollectdPlugin
from dbod_instances import DbodInstance, MalformedDbodInstance
import collectd  # pylint: disable=import-error


class DbodCollectdInnoDBClusterPingPlugin(DbodCollectdPlugin):

    """
    Plugin for the Ping metric.
    """
    log_dispatch = "| dbod_innodb_cluster_ping | {} | {:10} | [{}] | {}"

    def primary_check(self, instance):
        if instance.role == 'REPLICA':
            self.helper.update_instance_role(instance.id, 'PRIMARY')
            self.helper.write_to_cache(instance.db_name, 'role', 'PRIMARY')
            self.helper.write_to_cache(instance.db_name, 'role_id', 1)
            self.helper.update_instance_attribute(instance.db_name, 'slave', 'false')
            collectd.warning(
                self.log_dispatch.format(
                    "WARNING", instance.db_name, "",
                    "Instance role ({}) does not match the current role, updating role in dbod01 to PRIMARY".format(instance.role) ))
        try:
            seconds_since_epoch = int(time())
            collectd.debug("%s, name: %s"
                           % (type(instance).__name__, instance.db_name))
            ##############################
            # Connect to the database
            ##############################
            instance.connect()
            ##############################
            # Empty test table
            ##############################
            instance.execute(instance.ping_delete_query)
            ##############################
            # Insert test value
            ##############################
            collectd.debug("Insert ping values to: %s"
                           % instance.db_name)
            instance.insert(
                instance.ping_insert_query.format(seconds_since_epoch)
            )

            ####################################
            # Select the value we just inserted
            # to see if the select works
            ####################################
            collectd.debug("Select ping values from: %s"
                           % instance.db_name)
            value = instance.select(
                instance.ping_select_query)

            ####################################
            # Compare selected value and dispatch
            # 0 for success 1 for error
            ####################################
            if value == seconds_since_epoch:
                self.dispatch([0], instance.notification,
                              self.log_dispatch.format(
                                  "INFO", instance.db_name,
                                  "-1", "Notifications disabled"),
                              plugin_instance=instance.db_name)
                self.helper.update_state(instance,
                                         DbodInstance.RUNNING)
            else:
                # Connection was ok but insert or select failed
                collectd.error(
                    self.log_dispatch.format(
                        "ERROR", instance.db_name, "1",
                        "Inserted value different than selected."))
                self.dispatch([1], instance.notification,
                              self.log_dispatch.format(
                                  "ERROR", instance.db_name,
                                  "-1", "Notifications disabled"),
                              plugin_instance=instance.db_name)
                self.helper.update_state(instance, DbodInstance.BUSY)
        except Exception as exc:  # noqa
            """
            We catch all exceptions. No matter what kind of exception
            it was, we dispatch [1] and update instance status
            """  # pylint: disable=pointless-string-statement
            # https://twitter.com/gvanrossum/status/112670605505077248
            collectd.error(
                self.log_dispatch.format(
                    "ERROR", instance.db_name, "1", exc))
            self.dispatch([1], instance.notification,
                          self.log_dispatch.format(
                              "ERROR", instance.db_name,
                              "-1", "Notifications disabled"),
                          plugin_instance=instance.db_name)
            # we update the status there
            self.helper.update_state(instance, DbodInstance.STOPPED)
        finally:
            try:
                # Close connection to the database
                instance.disconnect()
            except Exception as exc:
                # We catch all exceptions - maybe the plugin
                # wasn't able to connect to the instance at all?
                collectd.warning(
                    self.log_dispatch.format(
                        "WARNING", instance.db_name, "",
                        "Not able to disconnect: %s" % exc))

    def replica_check(self, instance):
        if instance.role == 'PRIMARY':
            collectd.warning(
                self.log_dispatch.format(
                    "WARNING", instance.db_name, "",
                    "Instance role ({}) does not match the current role, updating role in dbod01 to REPLICA".format(instance.role)))
            self.helper.update_instance_role(instance.id, 'REPLICA')
            self.helper.write_to_cache(instance.db_name, 'role', 'REPLICA')
            self.helper.write_to_cache(instance.db_name, 'role_id', 2)
            self.helper.update_instance_attribute(instance.db_name, 'slave', 'true')
        try:
            collectd.debug("%s, name: %s"
                           % (type(instance).__name__, instance.db_name))
            ##############################
            # Connect to the database
            ##############################
            instance.connect()
            ####################################
            # Select the value we just inserted
            # to see if the select works
            ####################################
            collectd.debug("Select from d : %s"
                           % instance.db_name)
            value = instance.select(
                instance.ping_read_only_query)

            ####################################
            # Compare selected value and dispatch
            # 0 for success 1 for error
            ####################################
            collectd.debug("Comparing = selected: %s=%s"
                           % ('1', value))
            if value == 1:
                self.dispatch([0], instance.notification,
                              self.log_dispatch.format(
                                  "INFO", instance.db_name,
                                  "-1", "Notifications disabled"),
                              plugin_instance=instance.db_name)
                self.helper.update_state(instance,
                                         DbodInstance.RUNNING)
        except Exception as exc:
            """
            We catch all exceptions. No matter what kind of exception
            it was, we dispatch [1] and update instance status
            """  # pylint: disable=pointless-string-statement
            # https://twitter.com/gvanrossum/status/112670605505077248
            collectd.error(
                self.log_dispatch.format(
                    "ERROR", instance.db_name, "1", exc))
            self.dispatch([1], instance.notification,
                          self.log_dispatch.format(
                              "ERROR", instance.db_name,
                              "-1", "Notifications disabled"),
                          plugin_instance=instance.db_name)
            # we update the status there
            self.helper.update_state(instance, DbodInstance.STOPPED)
        finally:
            try:
                # Close connection to the database
                instance.disconnect()
            except Exception as exc:
                # We catch all exceptions - maybe the plugin
                # wasn't able to connect to the instance at all?
                collectd.warning(
                    self.log_dispatch.format(
                        "WARNING", instance.db_name, "",
                        "Not able to disconnect: %s" % exc))
    def read(self):
        """
        Read iterates over the instances running on a given host
            then does, for each instance:
                - connect()
                - insert() - test insert query
                - select() - test select query
                - delete() - test delete query

        Then we dispatch 0 or 1 depending on the success or failure
        """
        super(DbodCollectdInnoDBClusterPingPlugin, self).read()
        # Will fill this pattern with string.format

        # Get the time in seconds since the epoch
        collectd.debug("read() in DbodCollectdInnoDBClusterPingPlugin")
        for instance in self.instances:
            if isinstance(instance, MalformedDbodInstance):
                continue
            collectd.debug("-------")
            if instance.cluster_type != 'InnoDB':
                continue
            else:
                # Skip not active instances
                if not instance.active:
                    self.dispatch([-1], instance.notification,
                                  self.log_dispatch.format(
                                      "INFO", instance.db_name,
                                      "-1", "Instance not active."),
                                  plugin_instance=instance.db_name)
                    collectd.info(
                        self.log_dispatch.format(
                            "INFO", instance.db_name,
                            "-1", "Instance not active."))
                else:
                    try:
                        instance.connect()
                        read_only = instance.get_read_only_variable_value()
                        if read_only == 'ON':
                            self.replica_check(instance)
                        elif read_only == 'OFF':
                            self.primary_check(instance)
                    except Exception as ex:
                        collectd.error(
                            self.log_dispatch.format(
                                "ERROR", instance.db_name, "1", ex))
                        self.helper.update_state(instance, DbodInstance.STOPPED)
                        self.dispatch([1], instance.notification,
                                      self.log_dispatch.format(
                                          "ERROR", instance.db_name,
                                          "-1", "Notifications disabled"),
                                      plugin_instance=instance.db_name)
                        if instance.role == 'PRIMARY':
                            self.helper.update_instance_role(instance.id, 'REPLICA')
                            self.helper.write_to_cache(instance.db_name, 'role', 'REPLICA')
                            self.helper.write_to_cache(instance.db_name, 'role_id', 2)
                            self.helper.update_instance_attribute(instance.db_name, 'slave', 'true')

DbodCollectdInnoDBClusterPingPlugin()



import pytest
from mock import MagicMock
import dbod_instances
import dbod_plugin # noqa


@pytest.fixture
def plugin(mocker):
    """
    Mock instances etc
    """
    values = MagicMock()
    mocker.patch('collectd.Values', values)
    mocker.patch("dbod_instances.MySQLDbodInstance.__init__")
    mocker.patch("dbod_instances.PostgresDbodInstance.__init__")
    mocker.patch("dbod_instances.InfluxDbodInstance.__init__")
    mocker.patch("dbod_instances.MalformedDbodInstance.__init__")
    dbod_instances.MySQLDbodInstance.__init__.return_value = None
    dbod_instances.PostgresDbodInstance.__init__.return_value = None
    dbod_instances.InfluxDbodInstance.__init__.return_value = None
    dbod_instances.MalformedDbodInstance.__init__.return_value = None
    mocker.patch("dbod_plugin.DbodCollectdPlugin.__init__")
    mocker.patch("dbod_plugin.DbodCollectdPlugin.read")
    dbod_plugin.DbodCollectdPlugin.__init__.return_value = None


def test_is_subclass_of_dbod_plugin(plugin):
    """
    Test that DbodCollectdPluginEntitiesMalformed is a subclass of
        DbodCollectdPlugin
    """
    from dbod_entities_malformed \
        import DbodCollectdPluginEntitiesMalformed # noqa
    i = DbodCollectdPluginEntitiesMalformed()
    assert isinstance(i, dbod_plugin.DbodCollectdPlugin)


def test_read_calls_parent_read(mocker, plugin):
    """
    Parent class read must be called to refresh instances data
    """
    from dbod_entities_malformed \
        import DbodCollectdPluginEntitiesMalformed # noqa
    mocker.patch('dbod_plugin.DbodCollectdPlugin.read')
    i = DbodCollectdPluginEntitiesMalformed()
    i.instances = []
    i.collectd_type = "gauge"
    i.collectd_plugin = "name"
    i.read()
    dbod_plugin.DbodCollectdPlugin.read.assert_called()


@pytest.mark.parametrize(
    "classes,number",
    [
        ([dbod_instances.MySQLDbodInstance], 0),
        ([dbod_instances.MySQLDbodInstance,
          dbod_instances.MalformedDbodInstance], 1),
        ([dbod_instances.MySQLDbodInstance,
          dbod_instances.MalformedDbodInstance,
          dbod_instances.PostgresDbodInstance,
          dbod_instances.MalformedDbodInstance], 2),
        ([dbod_instances.MySQLDbodInstance,
          dbod_instances.MalformedDbodInstance,
          dbod_instances.PostgresDbodInstance,
          dbod_instances.InfluxDbodInstance,
          dbod_instances.MalformedDbodInstance,
          dbod_instances.MalformedDbodInstance], 3),
    ])
def test_read_goes_through_list_of_entities(plugin, mocker, classes, number):
    """
    Read should go through the list of instances, get number of
        MalformedDbodInstance, and call collectd dispatch
    """
    from dbod_entities_malformed \
        import DbodCollectdPluginEntitiesMalformed # noqa
    instances = []
    mocker.patch('dbod_plugin.DbodCollectdPlugin.read')
    for cls in classes:
        instance = cls()
        instance.state = "UNKNOWN"
        instances.append(instance)
    p = DbodCollectdPluginEntitiesMalformed()
    mocker.patch.object(p, "dispatch")
    mocker.patch.object(p, 'instances', instances, create=True)
    p.read()
    p.dispatch.assert_called_with([number])


def test_read_omits_instances_awaiting_approval(plugin, mocker):
    """
    Malformed instances with state AWAITING_APPROVAL should be ommited
    in the count
    """
    from dbod_entities_malformed \
        import DbodCollectdPluginEntitiesMalformed # noqa
    p = DbodCollectdPluginEntitiesMalformed()
    mocker.patch.object(p, 'dispatch')
    instance1 = dbod_instances.MalformedDbodInstance()
    instance1.state = "AWAITING_APPROVAL"
    instance2 = dbod_instances.MalformedDbodInstance()
    instance2.state = "UNKNOWN"
    p.instances = [instance1, instance2]
    p.read()
    p.dispatch.assert_called_with([1])



from mock import call
from dbod_plugin import DbodCollectdPlugin
from dbod_instances import DbodInstance, MalformedDbodInstance
from helpers import get_mocked_instance
from hypothesis import given, strategies as st, settings, HealthCheck
import mock
import pytest
patcher = mock.patch('dbod_plugin.DbodCollectdPlugin.__init__')
init = patcher.start()
init.return_value = None
from dbod_slave import DbodSlavePlugin # noqa
patcher.stop()


@pytest.fixture
def plugin(mocker):
    """
    Returns an instance of DbodCollectdPlugin with
    all the real things mocked:
        - plugin.helper mocked - the constructor
            opens the config files to read passwords etc.
        - plugin.read mocked - it calls helper.get_instances which
            sends a request DBOD API to get list of entities
    """
    # We mock init of the parent class because it calls collectd methods
    # to register read
    init = mocker.patch('dbod_plugin.DbodCollectdPlugin.__init__')
    init.return_value = None
    # Read from parent class would call DBOD API to get instances
    # running locally, we provide our own instances in the tests
    mocker.patch('dbod_plugin.DbodCollectdPlugin.read')
    mocker.patch('dbod_plugin.DbodCollectdPlugin.dispatch')
    from dbod_slave import DbodSlavePlugin
    p = DbodSlavePlugin()
    # Mock instances attribute - this noramlly comes from the DBOD API

    mocker.patch.object(p, 'helper', create=True)
    return p


def test_inheritance(mocker):
    """
    Test that DbodSlavePluign inherits from DbodCollectdPlugin
    and instance can be created (so it implements all the necessary methods)
    """
    init = mocker.patch('dbod_plugin.DbodCollectdPlugin.__init__')
    init.return_value = None
    s = DbodSlavePlugin()
    assert isinstance(s, DbodCollectdPlugin)


def test_read_calls_parent_read(mocker):
    """
    Test that DbodSlavePlugin.read() calls parent class read() method
    It has to be called to refresh instances from DBOD API
    """
    init = mocker.patch('dbod_plugin.DbodCollectdPlugin.__init__')
    read = mocker.patch('dbod_plugin.DbodCollectdPlugin.read')
    init.return_value = None
    s = DbodSlavePlugin()
    try:
        s.read()
    except Exception:
        pass
    read.assert_called()


def test_importing_calls_init():
    init.assert_called()


@given(
    instances=st.lists(
        # Create a list of 100 instance mocks
        st.builds(
            get_mocked_instance,
            db_name=st.integers(max_value=10),
            slave=st.booleans(),
            active=st.booleans(),
            malformed=st.booleans(),
            exception=st.sampled_from([None]),
            select_value=st.integers(
                min_value=1, max_value=10),
            notification=st.booleans()),
        min_size=1,
        max_size=20
        )
)
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_read_instance_methods(mocker, plugin, instances):
    """
    Test that depending on the instance attributes (active, slave)
    correct methods are called
    Additionally, test that MalformedDbodInstance is not taken
    into account at all
    We should perform any action only if instance is slave
    """
    #########################################
    # Mock various things
    #########################################
    # plugin.instances normally comes from DBOD API,
    # here we set it to mocks generated with hypothesis
    mocker.resetall()
    mocker.patch.object(plugin, 'instances', instances, create=True)
    ##################################################
    # The actual call that we're testing,
    # normally read is called by collectd!
    ##################################################
    plugin.read()
    # We iterate over mocked instances and check
    # that proper calls were made
    ##################################################
    for instance in instances:
        # Malformed instance shouldn't run anything
        if isinstance(instance, MalformedDbodInstance):
            instance.assert_not_called()
            continue
        # We should perform actions only on slave instances
        if not instance.slave:
            instance.assert_not_called()
            continue
        # We should connect and select from slave instances
        calls = [call.connect(),
                 call.get_replication_lag(),
                 call.disconnect()]
        instance.mock_calls == calls


@given(
    instances=st.lists(
        # Create a list of 100 instance mocks
        st.builds(
            get_mocked_instance,
            db_name=st.integers(max_value=10),
            slave=st.just(True),
            active=st.just(True),
            malformed=st.just(False),
            exception=st.sampled_from([None]),
            select_value=st.integers(
                min_value=1, max_value=10),
            # Simulate get_replicaiton_lag returning
            # from 1 to 3000 seconds
            get_replication_lag_value=st.integers(1, 10),
            max_lag_seconds=st.integers(1, 10),
            notification=st.booleans()
            ),
        min_size=1,
        max_size=20
        ),
)
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_read_dispatch(mocker, plugin, instances):
    """
    Test that depending on the get_replicaiton_lag()
    and max_lag_seconds we dispatch a correct value

    No exceptions tested here
    """
    #########################################
    # Mock various things
    #########################################
    # plugin.instances normally comes from DBOD API,
    # here we set it to mocks generated with hypothesis
    mocker.resetall()
    mocker.patch.object(plugin, 'instances', instances, create=True)
    ##################################################
    # The actual call that we're testing,
    # normally read is called by collectd!
    ##################################################
    plugin.dispatch.reset_mock()
    plugin.read()
    ##################################################
    # We collect ispatch calls here
    expected_calls = []
    # We iterate over mocked instances and check
    # that proper calls were made
    ##################################################
    log_dispatch = "| dbod_slave | {} | {:10} | [{}] | {}"

    for instance in instances:
        if instance.get_replication_lag() > instance.max_lag_seconds:
            expected_calls.append(call([1], instance.notification,
                log_dispatch.format("ERROR", instance.db_name,
                "-1", "Notifications disabled"),
                plugin_instance=instance.db_name))
        else:
            expected_calls.append(call([0], instance.notification,
                log_dispatch.format("INFO", instance.db_name,
                "-1", "Notifications disabled"),
                plugin_instance=instance.db_name))
    assert plugin.dispatch.mock_calls == expected_calls

@given(
    instances=st.lists(
        # Create a list of 100 instance mocks
        st.builds(
            get_mocked_instance,
            db_name=st.integers(max_value=10),
            slave=st.just(True),
            active=st.just(True),
            malformed=st.just(False),
            exception=st.sampled_from(
                ["connect", "disconnect", "get_replication_lag"]),
            select_value=st.integers(
                min_value=1, max_value=10),
            # Simulate get_replicaiton_lag returning
            # from 1 to 3000 seconds
            get_replication_lag_value=st.integers(1, 10),
            max_lag_seconds=st.integers(1, 10)
            ),
        min_size=1,
        max_size=20
        ),
)
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_read_dispatch_exceptions(mocker, plugin, instances):
    """
    Test that if any method raises exception we dispatch [1]
    """
    #########################################
    # Mock various things
    #########################################
    # plugin.instances normally comes from DBOD API,
    # here we set it to mocks generated with hypothesis
    mocker.resetall()
    mocker.patch.object(plugin, 'instances', instances, create=True)
    ##################################################
    # The actual call that we're testing,
    # normally read is called by collectd!
    ##################################################
    plugin.read()
    ##################################################
    # We collect ispatch calls here
    expected_calls = []
    # We iterate over mocked instances and check
    # that proper calls were made
    ##################################################
    log_dispatch = "| dbod_slave | {} | {:10} | [{}] | {}"

    for instance in instances:
        # If no exception raised or disconnect raised an exception
        if not instance.exception or instance.exception == "disconnect":
            if instance.get_replication_lag() > instance.max_lag_seconds:
                expected_calls.append(
                    call([1], instance.notification,
                        log_dispatch.format("ERROR", instance.db_name,
                        "-1", "Notifications disabled"),
                        plugin_instance=instance.db_name))
            else:
                expected_calls.append(
                    call([0], instance.notification,
                        log_dispatch.format("INFO", instance.db_name,
                        "-1", "Notifications disabled"),
                        plugin_instance=instance.db_name))
        elif instance.exception in ["connect", "get_replication_lag"]:
            expected_calls.append(
                call([1], instance.notification,
                    log_dispatch.format("ERROR", instance.db_name,
                    "-1", "Notifications disabled"),
                    plugin_instance=instance.db_name))
    assert plugin.dispatch.mock_calls == expected_calls
    mocker.resetall()


@given(
    instances=st.lists(
        # Create a list of 100 instance mocks
        st.builds(
            get_mocked_instance,
            db_name=st.integers(max_value=10),
            slave=st.just(True),
            active=st.just(True),
            malformed=st.just(False),
            exception=st.sampled_from(
                ["connect", "get_replication_lag", "disconnect", None]),
            select_value=st.integers(
                min_value=1, max_value=10),
            # Simulate get_replicaiton_lag returning
            # from 1 to 3000 seconds
            get_replication_lag_value=st.integers(1, 10),
            max_lag_seconds=st.integers(1, 10)
            ),
        min_size=1,
        max_size=20
        ),
)
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_read_update_state(mocker, plugin, instances):
    """
    Test that we call properly update_state, to update
    instance state in DBOD API
    """
    #########################################
    # Mock various things
    #########################################
    # plugin.instances normally comes from DBOD API,
    # here we set it to mocks generated with hypothesis
    mocker.resetall()
    mocker.patch.object(plugin, 'instances', instances, create=True)
    ##################################################
    # The actual call that we're testing,
    # normally read is called by collectd!
    ##################################################
    plugin.helper.update_state.reset_mock()
    plugin.read()
    ##################################################
    # We collect ispatch calls here
    expected_calls = []
    # We iterate over mocked instances and check
    # that proper calls were made
    ##################################################
    for instance in instances:
        # If exception was raised on connection, or getting the lag
        # update instance to stopped
        if instance.exception and \
                instance.exception in ["connect", "get_replication_lag"]:
            expected_calls.append(call(instance, DbodInstance.STOPPED))
            continue
        if instance.get_replication_lag() > instance.max_lag_seconds:
            expected_calls.append(call(instance, DbodInstance.BUSY))
        else:
            expected_calls.append(call(instance, DbodInstance.RUNNING))
    assert plugin.helper.update_state.mock_calls == expected_calls


@given(
    instances=st.lists(
        # Create a list of 100 instance mocks
        st.builds(
            get_mocked_instance,
            db_name=st.integers(0, 4),
            slave=st.just(True),
            active=st.just(True),
            malformed=st.just(False),
            exception=st.sampled_from(
                ["connect", "get_replication_lag", "disconnect", None]),
            select_value=st.integers(
                min_value=1, max_value=5),
            # Simulate get_replicaiton_lag returning
            # from 1 to 3000 seconds
            get_replication_lag_value=st.integers(1, 10),
            max_lag_seconds=st.integers(1, 10),
            notification=st.booleans()
            ),
        min_size=1,
        max_size=20
        ),
)
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_read_logging(mocker, plugin, instances):
    """
    Test that messages are logged (errors) on dispatch
    """
    import collectd
    collectd.reset_mock()
    mocker.resetall()
    mocker.patch.object(plugin, 'instances', instances, create=True)
    plugin.helper.update_state.reset_mock()
    plugin.read()
    expected_calls_warning = []
    expected_calls_info = []
    expected_calls_error = []

    # We iterate over mocked instances and check
    # that proper calls to collectd.error() were made
    ##################################################
    log_dispatch = "| dbod_slave | {} | {:10} | [{}] | {}"
    #############
    for instance in instances:
        # If exception was raised on connection, or getting the lag
        # update instance to stopped
        if instance.exception and \
                instance.exception in ["connect", "get_replication_lag"]:
            expected_calls_error.append(
                call(log_dispatch.format(
                        "ERROR", instance.db_name,
                        "1", "Exception: %s" % instance.exception_message)))
            continue
        if instance.get_replication_lag() > instance.max_lag_seconds:
            expected_calls_error.append(
                call(log_dispatch.format(
                        "ERROR", instance.db_name,
                        "1", "Instance lags %s seconds behind master"
                        % instance.get_replication_lag_value)))
        if instance.exception == "disconnect":
            # log disconnect errors as debug
            expected_calls_warning.append(
                call(log_dispatch.format(
                        "WARNING", instance.db_name,
                        "", "Not able to disconnect: %s"
                        % instance.exception_message)))

    assert collectd.error.mock_calls == expected_calls_error
    assert collectd.info.mock_calls == expected_calls_info
    assert collectd.warning.mock_calls == expected_calls_warning



"""
DBOD ping
"""

from time import time
from dbod_plugin import DbodCollectdPlugin
from dbod_instances import DbodInstance, MalformedDbodInstance
import collectd  # pylint: disable=import-error


class DbodCollectdPluginPing(DbodCollectdPlugin):
    """
    Plugin for the Ping metric. What it does:

    """
    log_dispatch = "| dbod_ping | {} | {:10} | [{}] | {}"

    def read(self):
        """
        Read iterates over the instances running on a given host
            then does, for each instance:
                - connect()
                - create_table()
                - insert() - test insert query
                - select() - test select query

        Then we dispatch 0 or 1 depending on the success or failure
        """
        super(DbodCollectdPluginPing, self).read()
        # Will fill this pattern with string.format

        # Get the time in seconds since the epoch
        collectd.debug("read() in DbodCollectdPluginPing")
        seconds_since_epoch = int(time())

        for instance in self.instances:
            if isinstance(instance, MalformedDbodInstance):
                continue

            collectd.debug("-------")

            # Skip not active instances
            if not instance.active:
                self.dispatch([-1], instance.notification,
                              self.log_dispatch.format(
                                  "INFO", instance.db_name,
                                  "-1", "Instance not active."),
                              plugin_instance=instance.db_name)
                collectd.info(
                    self.log_dispatch.format(
                        "INFO", instance.db_name,
                        "-1", "Instance not active."))
                continue

            # Skip slave instances
            if instance.slave:
                continue
            elif instance.cluster_type == 'InnoDB':
                continue
            try:
                collectd.debug("%s, name: %s"
                               % (type(instance).__name__, instance.db_name))
                # Iterate over all instances
                ##############################
                # Connect to the database
                ##############################
                instance.connect()

                ##############################
                # Create test table
                ##############################
                instance.create(instance.ping_create_query)
                ##############################
                # Empty test table
                ##############################
                instance.execute(instance.ping_delete_query)
                ##############################
                # Insert test value
                ##############################
                collectd.debug("Insert ping values to: %s"
                               % instance.db_name)
                instance.insert(
                    instance.ping_insert_query.format(seconds_since_epoch)
                )

                ####################################
                # Select the value we just inserted
                # to see if the select works
                ####################################
                collectd.debug("Select ping values from: %s"
                               % instance.db_name)
                value = instance.select(
                    instance.ping_select_query)

                ####################################
                # Compare selected value and dispatch
                # 0 for success 1 for error
                ####################################
                collectd.debug("Comparing inserted=selected: %s=%s"
                               % (seconds_since_epoch, value))
                if value == seconds_since_epoch:
                    self.dispatch([0], instance.notification,
                                  self.log_dispatch.format(
                                      "INFO", instance.db_name,
                                      "-1", "Notifications disabled"),
                                  plugin_instance=instance.db_name)
                    self.helper.update_state(instance,
                                             DbodInstance.RUNNING)
                else:
                    # Connection was ok but insert or select failed
                    collectd.error(
                        self.log_dispatch.format(
                            "ERROR", instance.db_name, "1",
                            "Inserted value different than selected."))
                    self.dispatch([1], instance.notification,
                                  self.log_dispatch.format(
                                      "ERROR", instance.db_name,
                                      "-1", "Notifications disabled"),
                                  plugin_instance=instance.db_name)
                    self.helper.update_state(instance, DbodInstance.BUSY)
            except Exception as exc: # noqa
                """
                We catch all exceptions. No matter what kind of exception
                it was, we dispatch [1] and update instance status
                """  # pylint: disable=pointless-string-statement
                # https://twitter.com/gvanrossum/status/112670605505077248
                collectd.error(
                    self.log_dispatch.format(
                        "ERROR", instance.db_name, "1", exc))
                self.dispatch([1], instance.notification,
                              self.log_dispatch.format(
                                  "ERROR", instance.db_name,
                                  "-1", "Notifications disabled"),
                              plugin_instance=instance.db_name)
                # we update the status there
                self.helper.update_state(instance, DbodInstance.STOPPED)
            finally:
                try:
                    # Close connection to the database
                    instance.disconnect()
                except Exception as exc:
                    # We catch all exceptions - maybe the plugin
                    # wasn't able to connect to the instance at all?
                    collectd.warning(
                        self.log_dispatch.format(
                            "WARNING", instance.db_name, "",
                            "Not able to disconnect: %s" % exc))


DbodCollectdPluginPing()



"""
DBOD entities malformed
"""

from dbod_plugin import DbodCollectdPlugin
from dbod_instances import MalformedDbodInstance


class DbodCollectdPluginEntitiesMalformed(DbodCollectdPlugin):
    """
    Entities malformed metric:

    """
    def read(self):
        """
        Read number of malformed dbod entities
        """
        super(DbodCollectdPluginEntitiesMalformed, self).read()
        number_of_malformed = 0
        for instance in self.instances:
            if isinstance(instance, MalformedDbodInstance):
                if instance.state == "AWAITING_APPROVAL":
                    continue
                number_of_malformed += 1
        self.dispatch([number_of_malformed])


DbodCollectdPluginEntitiesMalformed()



"""
DBOD replication
"""

from dbod_plugin import DbodCollectdPlugin
from dbod_instances import DbodInstance, MalformedDbodInstance
import collectd  # pylint: disable=import-error


class DbodReplicationPlugin(DbodCollectdPlugin):
    """
    DBOD replication plugin
    """

    log_dispatch = "| dbod_replication | {} | {:10} | [{}] | {}"

    def mysql_repl_threads_check(self, instance):
        collectd.debug(" (MySQL) dbod_replication - checking MySQL replica {} replication threads".format(instance.db_name))
        try:
            instance.connect()
            io_thread_status = instance.get_replica_io_thread_status()
            sql_thread_status = instance.get_replica_sql_thread_status()
            if io_thread_status != 'Yes' or sql_thread_status != 'Yes':
                collectd.error(
                    self.log_dispatch.format(
                        "ERROR", instance.db_name, "1",
                        "(MySQL) One of the replication threads is not running; io_thread_running: {}, sql_thread_running: {}".format(
                            io_thread_status, sql_thread_status)))
                self.dispatch([1], instance.notification,
                              self.log_dispatch.format(
                                  "ERROR", instance.db_name,
                                  "-1", "Notifications disabled"),
                              plugin_instance=instance.db_name)
                self.helper.update_state(instance, DbodInstance.BUSY)
            else:
                self.dispatch([0], instance.notification,
                              self.log_dispatch.format(
                                  "INFO", instance.db_name,
                                  "-1", "Notifications disabled"),
                              plugin_instance=instance.db_name)
                self.helper.update_state(instance,
                                         DbodInstance.RUNNING)
        except Exception as exc:
            # Catch all exception and dispatch error
            collectd.error(
                self.log_dispatch.format(
                    "ERROR", instance.db_name, "1",
                    "Exception: {}".format(exc)))
            self.dispatch([1], instance.notification,
                          self.log_dispatch.format(
                              "ERROR", instance.db_name,
                              "-1", "Notifications disabled"),
                          plugin_instance=instance.db_name)
            self.helper.update_state(instance, DbodInstance.STOPPED)
        finally:
            try:
                # Close connection to the database
                instance.disconnect()
            except Exception as exc:
                # We catch all exceptions - maybe the plugin
                # wasn't able to connect to the instance at all?
                collectd.warning(
                    self.log_dispatch.format(
                        "WARNING", instance.db_name, "",
                        "Not able to disconnect: {}".format(exc)))

    def pg_repl_check(self,instance):
        try:
            instance.connect()
            repl_status = instance.get_replication_status()
            if repl_status != 'streaming':
                collectd.error(
                    self.log_dispatch.format(
                        "ERROR", instance.db_name, "1",
                        "(PostgreSQL) Replication is broken, replication status is : {}".format(
                            repl_status)))
                self.dispatch([1], instance.notification,
                              self.log_dispatch.format(
                                  "ERROR", instance.db_name,
                                  "-1", "Notifications disabled"),
                              plugin_instance=instance.db_name)
                self.helper.update_state(instance, DbodInstance.BUSY)
            else:
                self.dispatch([0], instance.notification,
                              self.log_dispatch.format(
                                  "INFO", instance.db_name,
                                  "-1", "Notifications disabled"),
                              plugin_instance=instance.db_name)
                self.helper.update_state(instance,
                                         DbodInstance.RUNNING)
        except Exception as exc:
            # Catch all exception and dispatch error
            collectd.error(
                self.log_dispatch.format(
                    "ERROR", instance.db_name, "1",
                    "Exception: {}".format(exc)))
            self.dispatch([1], instance.notification,
                          self.log_dispatch.format(
                              "ERROR", instance.db_name,
                              "-1", "Notifications disabled"),
                          plugin_instance=instance.db_name)
            self.helper.update_state(instance, DbodInstance.STOPPED)
        finally:
            try:
                # Close connection to the database
                instance.disconnect()
            except Exception as exc:
                # We catch all exceptions - maybe the plugin
                # wasn't able to connect to the instance at all?
                collectd.warning(
                    self.log_dispatch.format(
                        "WARNING", instance.db_name, "",
                        "Not able to disconnect: {}".format(exc)))

    def read(self):
        """
        Call super - this will refresh list of instances from the Dbod API
        Then iterate over the list of instances, and check the replication status if they are replicas
        """
        super(DbodReplicationPlugin, self).read()
        for instance in self.instances:
            # MalformedDbodInstance means broken data in DBOD API
            # this is handled by another collectd plugin
            # Catch all exception and dispatch error
            if isinstance(instance, MalformedDbodInstance):
                continue
            if instance.cluster_type == 'InnoDB':
                continue
            if instance.slave:

                collectd.debug("dbod_replication - instance {} is a replica".format(instance.db_name))
                if instance.db_type == 'MYSQL':
                    self.mysql_repl_threads_check(instance)
                elif instance.db_type == 'PG':
                    self.pg_repl_check(instance)




DbodReplicationPlugin()



#!/usr/bin/python3

from dbod_helpers import DbodCollectdHelper, \
    DbodConfigException
from mock import MagicMock, call
from pytest import raises
import pytest
import requests
import json
import dbod_helpers
import dbod_instances
import yaml
import platform
PYTHON_VERSION = platform.python_version()
from pkg_resources import parse_version as version
import cerndb
from  pathlib import Path
import builtins

@pytest.fixture
def helper_mocks(mocker):
    """
    Mock all external libraries calls
    """

    mocker.patch('pathlib.Path.exists')
    path_mock = MagicMock()
    Path.exists = path_mock
    #Patch https://gitlab.cern.ch/db/python3-cerndb/-/blob/master/cerndb/config/__init__.py#L41
    path_mock.return_value = True
    mocker.patch('builtins.open')
    open_mock = MagicMock()
    builtins.open = open_mock
    #Patch https://gitlab.cern.ch/db/python3-cerndb/-/blob/master/cerndb/config/__init__.py#L52
    open_mock.return_value.__enter__.return_value = ('''
            api:
              cachefile: /some/cache/file.json
              host: https://some-url.cern.ch
              entity_endpoint: api/v1/instance
              cluster_member_endpoint: qa/api/v1/cluster-member
              user: dbod-api
              password: pass-api
            mysql:
              user: uu
              password: uu
            dbmon:
              user: uu
              password: uu
            dbod_database:
              db_name: itcore
              db_account: DBONDEMAND
              db_pwd: 12131
            influx2:
              client: /path/to/influx/client
            ''')
    mocker.patch('dbod_helpers.gethostname', create=True)
    dbod_helpers.gethostname.return_value = "dbod-test1.cern.ch"
    # Mock requests
    mocker.patch("requests.get")
    requests.get.return_value.status_code = 200
    # Mock instance
    mocker.patch('dbod_instances.DbodInstance.__init__')
    mocker.patch('dbod_instances.MySQLDbodInstance.__init__')
    mocker.patch('dbod_instances.PostgresDbodInstance.__init__')
    mocker.patch('dbod_instances.InfluxDbodInstance.__init__')
    mocker.patch('dbod_instances.MalformedDbodInstance.__init__')
    dbod_instances.MySQLDbodInstance.__init__.return_value = None
    dbod_instances.PostgresDbodInstance.__init__.return_value = None
    dbod_instances.InfluxDbodInstance.__init__.return_value = None
    dbod_instances.MalformedDbodInstance.__init__.return_value = None
    dbod_instances.DbodInstance.__init__.return_value = None

    return {"open_mock": open_mock}


def test_dbod_helper_reads_config(mocker, helper_mocks):
    """
    Test a method to get the "source of truth"
    """
    # Test configuration was read
    h = DbodCollectdHelper()
    assert h.cachefile == "/some/cache/file.json"
    assert h.api_url == 'https://some-url.cern.ch'
    assert h.api_user == 'dbod-api'
    assert h.api_password == 'pass-api'
    assert h.user == 'uu'



def test_dbod_helper_reads_config_malformed(mocker, helper_mocks):
    """
    Test a method to get the "source of truth"
    May it be YAML from the disk
    """
    # Test configuration was read
    helper_mocks["open_mock"].return_value.__enter__.return_value  = ('''a''')
    with raises(DbodConfigException) as exc:
        DbodCollectdHelper()
    assert str(exc.value) == "Error in DBOD config file (config.yaml):" \
                             " missing key 'api'"
    ##########################################################
    helper_mocks["open_mock"].return_value.__enter__.return_value = ('''
            api:
            ''')
    with raises(DbodConfigException) as exc:
        DbodCollectdHelper()
    assert str(exc.value) == "Error in DBOD config file (config.yaml): missing nested key"

    ##########################################################

    helper_mocks["open_mock"].return_value.__enter__.return_value = ('''
            api:
              cachefile: /some/cache/file.json
              host: https://some-url.cern.ch
            ''')
    with raises(DbodConfigException) as exc:
        DbodCollectdHelper()
    assert str(exc.value) == "Error in DBOD config file (config.yaml): " \
                             "missing key 'entity_endpoint'"
    ##########################################################
    helper_mocks["open_mock"].return_value.__enter__.side_effect =  FileNotFoundError("sth wrong")
    with raises(DbodConfigException) as exc:
        DbodCollectdHelper()
    assert str(exc.value) == "sth wrong"


@pytest.fixture
def instances_from_json(helper_mocks, mocker):
    requests.get.return_value.status_code = 404

    mocker.patch('dbod_helpers.open')
    mocker.patch('json.load')
    json.load.return_value = [
            {"instance": "1", "type": "PG", "state": "RUNNING"},
            {"instance": "2", "type": "MYSQL", "state": "RUNNING"},
            {"instance": "3", "type": "InfluxDB", "state": "RUNNING"}
        ]

    open_file = MagicMock()
    dbod_helpers.open.return_value.__enter__.return_value = open_file

def test_get_instances_from_json(mocker, tmpdir, instances_from_json):
    """
    use entities.json
    """

    h = DbodCollectdHelper()
    instances = h.get_instances()
    json.load.assert_called_with(
        dbod_helpers.open.return_value.__enter__.return_value)
    dbod_helpers.open.assert_called_with(h.cachefile)
    assert dbod_instances.MySQLDbodInstance.__init__. \
        mock_calls == [call({"instance": "2", "type": "MYSQL", "state": "RUNNING"},
                            influx2_client="/path/to/influx/client", user="uu", password="uu")]
    assert dbod_instances.PostgresDbodInstance.__init__. \
        mock_calls == [call({"instance": "1", "type": "PG", "state": "RUNNING"},
                            influx2_client="/path/to/influx/client", user="uu", password="uu")]
    assert dbod_instances.InfluxDbodInstance.__init__. \
        mock_calls == [call({"instance": "3", "type": "InfluxDB", "state": "RUNNING"},
                            influx2_client="/path/to/influx/client", user="uu", password="uu")]
    assert len(instances) == 3

def test_get_instances_instance_raises_exception(
        mocker, tmpdir, instances_from_json):
    """
    When DbodInstance creation raises exception - we should get
        MalformedDbodInstance
    """
    dbod_instances.MySQLDbodInstance.__init__.side_effect = \
        dbod_helpers.EntityMalformedException("sth wrong")
    h = DbodCollectdHelper()
    instances = h.get_instances()
    assert dbod_instances.MalformedDbodInstance.__init__. \
        mock_calls == [call({"instance": "2", "type": "MYSQL", "state": "RUNNING"},
                            user="uu", password="uu")]
    assert len(instances) == 3
    assert isinstance(instances[1], dbod_instances.MalformedDbodInstance)

def test_get_instances_request_raises_exception_use_json(
        mocker, tmpdir, instances_from_json):
    requests.get.side_effect = requests.ConnectionError("Something wrong")
    h = DbodCollectdHelper()
    instances = h.get_instances()
    json.load.assert_called_with(
        dbod_helpers.open.return_value.__enter__.return_value)
    assert len(instances) == 3



def test_get_instances_type_missing(mocker, tmpdir, instances_from_json):
    """
    Test get_instances reading entities.json and failing with type missing
        Then it creates a list of DbodInstance object
        which is returned
    """
    mocker.patch('dbod_helpers.open')
    mocker.patch('json.load')
    json.load.return_value = [
            {"instance": "1", "type": "MYSQL", "state": "RUNNING"},
            {"instance": "2", "type": "InfluxDB", "state": "RUNNING"},
            {"instance": "3", "type": "MYSQL", "state": "RUNNING"},
            {"instance": "4", "state": "RUNNING"}
        ]

    open_file = MagicMock()
    dbod_helpers.open.return_value.__enter__.return_value = open_file
    h = DbodCollectdHelper()
    instances = h.get_instances()
    assert len(instances) == 4
    assert dbod_instances.MalformedDbodInstance.__init__. \
        mock_calls == [call({"instance": "4", "state": "RUNNING"},
                            password='uu', user='uu')]
    assert isinstance(instances[3], dbod_instances.MalformedDbodInstance)


@pytest.mark.parametrize(
    "state,api_url,endpoint,id,status_code,current_state",
    [("STOPPED", "api.cern.ch:3", "api/test", "db1",
      204, "RUNNING"),
     ("RUNNING", "a.c/232", "api/ent", "db2",
      204, "RUNNING"),
     ("BUSY", "test-api:33", "api/entities", "db3",
      200, "RUNNING"),
     ("RUNNING", "test-api:33", "api/entities", "db5",
      200, "AWAITING_APPROVAL"),
     ("BUSY", "test-api:33", "api/entities", "db4",
      200, "MAINTENANCE"),
     ])
def test_update_state(mocker, state, api_url, endpoint, id,
                       status_code, current_state,test_entities_file_name):
    init = mocker.patch('dbod_helpers.DbodCollectdHelper.__init__')
    init.return_value = None
    helper = DbodCollectdHelper()
    helper.api_url = api_url
    helper.instance_endpoint = endpoint
    helper.api_user = "user"
    helper.api_password = "pass"
    helper.cachefile = test_entities_file_name

    import requests
    import collectd
    mocker.resetall()
    requests.put.reset_mock()

    class Instance:
        pass
    instance = Instance()
    instance.id = id
    instance.db_name = id
    instance.state = current_state
    expected_url = "%s/%s/%s" % (api_url, endpoint, id)
    requests.put.return_value.status_code = status_code
    helper.update_state(instance, state)

    if current_state in ["AWAITING_APPROVAL", "MAINTENANCE"]:
        requests.put.assert_not_called()
        collectd.debug.assert_called_with(
            "Not updating state of %s. Instance state: %s"
            % (instance.db_name, current_state))
    elif current_state == state:
        requests.put.assert_not_called()
        collectd.debug("Instance %s state [%s] has not changed"
                        % (instance.db_name, instance.state))
    else:
        collectd.debug("Updating %s instance state to: %s"
                        % (instance.db_name, instance.state))
        expected_url = "%s/%s/%s" % (api_url,
                                    endpoint,
                                    instance.id)

        data_to_post = {"state": state}
        headers={'auth': '{ "admin" : true , "groups" : [ ] , "owner" : "collectd" }'}
        requests.put.assert_called_with(
            expected_url,
            headers=headers,
            auth=(helper.api_user, helper.api_password),
            json=data_to_post,
            verify=False
            )
        if status_code != 204:
            collectd.error.assert_called_with(
                "Failed updating state of %s. Response code %s, url %s, self.api_url %s, self.instance_endpoint %s, headers %s"
                % (instance.db_name, status_code, expected_url, api_url, endpoint, headers))




"""
DBOD plugin
"""

from abc import ABCMeta, abstractmethod
from dbod_helpers import DbodCollectdHelper
import collectd  # pylint: disable=import-error


# compatible with Python 2 *and* 3:
ABC = ABCMeta('ABC', (object,), {'__slots__': ()})


class DbodCollectdPlugin(ABC):
    """
    DBOD collectd plugin
    """
    def __init__(self):
        """
        Register collectd read and configure
        """
        self.instances = []
        collectd.debug("Registering init and read")
        collectd.register_init(self.collectd_init)
        collectd.register_read(self.read)
        # Set type and name to be dispatched
        self.collectd_type = "gauge"
        # self.collectd_plugin = \
        #    "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
        self.collectd_plugin = self.__class__.__module__
        collectd.debug("Creating DbodCollectdHelper")
        self.helper = DbodCollectdHelper()

    def collectd_init(self):
        """
        In this method we'll get all the running DBOD instances
        from the current host using DbodCollectdHelper,
        it's registered as init method with collectd using register_init
        """
        self.instances = self.helper.get_instances()

    @abstractmethod
    def read(self):
        """
        A method to actually read something useful
        and dispatch values, must be implemented in the actual plugins
        In the parent method we refresh the instances
        """
        self.instances = self.helper.get_instances()

    def dispatch(self, values, notification=True, log_dispatch="",
                 plugin_instance=None):
        """
        Dispatch values to collectd
        """
        metric = collectd.Values()
        metric.type = self.collectd_type
        metric.plugin = self.collectd_plugin

        if not notification:
            values = [-1]
            collectd.info(log_dispatch)

        collectd.debug("Dispatch plugin: %s, value: %s" %
                       (metric.plugin, values))

        if plugin_instance:
            metric.plugin_instance = plugin_instance

        metric.dispatch(values=values)



"""
DBOD instances
"""

from distutils.util import strtobool
from abc import ABCMeta, abstractmethod
import requests
import psycopg2
import pymysql
import subprocess
from distutils.version import StrictVersion

try:
    from itertools import izip as zip  # pylint: disable=redefined-builtin
except ImportError:  # Python 3.x
    pass


# compatible with Python 2 *and* 3:
ABC = ABCMeta('ABC', (object,), {'__slots__': ()})


# Python 2 - 3 compatibility
try:
    basestring
except NameError:
    basestring = (str, bytes)  # pylint: disable=invalid-name


class EntityMalformedException(Exception):
    """
    Raised when there is missing information in the metadata
    returned by the DBOD API
    """


class DbodInstance(ABC):
    # pylint: disable=too-many-instance-attributes
    # All of the attributes are considered necessary

    """
    Represents DBOD instance and is created from dictionary
        returned by a DBOD API call
    """
    STOPPED = "STOPPED"
    RUNNING = "RUNNING"
    BUSY = "BUSY"

    def __init__(self, api_dict, user=None, password=None, influx2_client=None):
        """
        Takes dictionary as it was returned by the DBOD API
            and fills instance attributes
        user - since user/password doesn't come from the API information
            we nedd separate args to set them

        Raises:
            EntityMalformedException - when one of the necessary attributes is
                not present in the dictionary.
        """
        self.connection = None
        if user:
            self.user = user
        if password:
            self.password = password
        if influx2_client:
            self.influx2_client = influx2_client

        try:
            self.instance = "dbod-%s.cern.ch" \
                % (api_dict["db_name"].replace("_", "-"))
            #check if we use the old api:
            if "db_type" in api_dict:
                self.db_type = api_dict["db_type"]
                self.active = api_dict["active"]
            else:
                self.id = api_dict["id"]
                self.db_type = api_dict["type"]
                self.active = (api_dict["status"] == "ACTIVE")
            self.port = api_dict["port"]
            self.socket = api_dict["socket"]
            self.category = api_dict["class"]
            self.notification = api_dict["attributes"]["notifications"]
            self.state = api_dict["state"]
            self.db_name = api_dict["db_name"]
            self.db_version = api_dict["version"]
            self.bindir = api_dict["bindir"]
            self.cluster_type = api_dict['cluster_type']
            self.role = api_dict['role']
        except KeyError as e:
            raise EntityMalformedException(
                "Attribute %s missing from the metadata! (DBOD API)"
                % e.args[0])

        #####################################
        # Change attribute types
        #####################################
        try:
            self.port = int(self.port)
        except ValueError as e:
            raise EntityMalformedException(
                "Attribute port has incorrect value! (DBOD API metadata): %s"
                % (e))
        if isinstance(self.notification, basestring):
            try:
                self.notification = bool(strtobool(self.notification))
            except ValueError as e:
                raise EntityMalformedException(
                    "Attribute notifications has incorrect value! "
                    "(DBOD API metadata): %s" % (e))

        ########################################
        # If slave was not defined, use False
        ########################################
        try:
            self.slave = api_dict["attributes"]["slave"]
            if isinstance(self.slave, basestring):
                try:
                    self.slave = bool(strtobool(self.slave))
                except ValueError as e:
                    raise EntityMalformedException(
                        "Attribute slave has incorrect value! "
                        "(DBOD API metadata): %s" % (e))
        except KeyError:
            self.slave = False

        #############################################################
        # If slave monit_notif_max_lag_seconds is not defined, use 0
        #############################################################
        try:
            self.max_lag_seconds = \
                int(api_dict["attributes"]["monit_notif_max_lag_seconds"])
        except KeyError:
            self.max_lag_seconds = 300

    @abstractmethod
    def connect(self):
        """Connect to an instance."""

    def disconnect(self):
        """
        Default implementation that works for both
            Postgres and MySQL library
        """
        self.connection.close()

    def execute(self, statement):
        """
        Default implementation that works for both
            Postgres and MySQL library
        """
        cursor = self.connection.cursor()
        cursor.execute(statement)
        self.connection.commit()
        cursor.close()

    def select(self, statement):
        """
        Default implementation that works for both
            Postgres and MySQL library
        For now it can select only one value!
        """
        cursor = self.connection.cursor()
        cursor.execute(statement)
        results = cursor.fetchall()
        cursor.close()
        return results[0][0]

    def insert(self, query):
        """
        By default it just executes, maybe for InfluxDB
        we need to override it, to be seen...
        """
        self.execute(query)

    def create(self, query):
        """
        Same as 'insert'. To be verified if needed at all
        """
        self.execute(query)

    def get_replication_lag(self):
        """
        Unfortunately we need a separate function for that.
        This is because while in Postgres, we can get this info with
        a select statement, in mysql we have to run 'SHOW SLAVE STATUS'.
        Show slave status, doesn't allow for easy filtering for the info that
        we're interested with, so we have to do it in Python.

        Implementation in the base class is the one that will be used in
        Postgres. Also, future release (8.0) of mysql has a table
        that can be selected for this info, then we can remove this method
        and use plain select
        """
        return self.select(self.slave_lag_query)  # pylint: disable=no-member


class MySQLDbodInstance(DbodInstance):
    """Represents a MySQL instance."""
    ping_insert_query = "INSERT INTO collectd_ping VALUES ({0});"
    ping_select_query = "SELECT * FROM collectd_ping;"
    ping_create_query = None
    ping_delete_query = "DELETE FROM collectd_ping;"
    slave_lag_query = "SHOW SLAVE STATUS"
    read_only_query = "SHOW VARIABLES LIKE 'read_only'"
    replica_status_query = "SHOW REPLICA STATUS"
    ping_read_only_query = "SELECT  1 FROM DUAL"
    def connect(self):
        self.connection = pymysql.connect(
            unix_socket=self.socket,
            user=self.user,
            passwd=self.password,
            db="dod_dbmon",
            port=self.port,
            connect_timeout=10)

    def create(self, query):
        pass

    def get_replication_lag(self):
        """
        SHOW SLAVE STATUS, then we have to process what we get
        from this statement to get the interesting bit - Seconds_Behind_Master
        """
        cursor = self.connection.cursor()
        cursor.execute(self.slave_lag_query)
        results = cursor.fetchone()
        self.connection.commit()
        cursor.close()
        column_names = [desc[0] for desc in cursor.description]
        results = dict(zip(column_names, results))
        return results['Seconds_Behind_Master']

    def get_read_only_variable_value(self):
        """
        executes show variables like 'read_only'
        """
        cursor = self.connection.cursor()
        cursor.execute(self.read_only_query)
        results = cursor.fetchall()
        self.connection.commit()
        cursor.close()
        dct = dict(results)
        return dct['read_only']

    def get_replica_io_thread_status(self):
        """
        executes SHOW SLAVE STATUS or SHOW REPLICA STATUS depending on the MySQL server version
        and checks the status of the Replica IO thread
        """
        result_field = ''
        cursor = self.connection.cursor()
        if StrictVersion(self.db_version) > StrictVersion('8.0.0'):
            cursor.execute(self.replica_status_query)
            result_field = 'Replica_IO_Running'
        elif StrictVersion(self.db_version) < StrictVersion('8.0.0'):
            cursor.execute(self.slave_lag_query)
            result_field = 'Slave_IO_Running'
        results = cursor.fetchone()
        self.connection.commit()
        cursor.close()
        column_names = [desc[0] for desc in cursor.description]
        results = dict(zip(column_names, results))
        return results[result_field]

    def get_replica_sql_thread_status(self):
        """
        executes SHOW SLAVE STATUS or SHOW REPLICA STATUS depending on the MySQL server version
        and checks the status of the Replica SQL thread
        """
        result_field = ''
        cursor = self.connection.cursor()
        if StrictVersion(self.db_version) > StrictVersion('8.0.0'):
            cursor.execute(self.replica_status_query)
            result_field = 'Replica_SQL_Running'
        elif StrictVersion(self.db_version) < StrictVersion('8.0.0'):
            cursor.execute(self.slave_lag_query)
            result_field = 'Slave_SQL_Running'
        results = cursor.fetchone()
        self.connection.commit()
        cursor.close()
        column_names = [desc[0] for desc in cursor.description]
        results = dict(zip(column_names, results))
        return results[result_field]

class PostgresDbodInstance(DbodInstance):
    """Represents a Postgresql instance. Is a subclass of DbodInstance"""
    ping_insert_query = "INSERT INTO collectd_ping VALUES ({0});"
    ping_select_query = "SELECT * FROM collectd_ping;"
    ping_create_query = None
    ping_delete_query = "DELETE FROM collectd_ping;"
    slave_lag_query = "SELECT CASE WHEN pg_last_wal_receive_lsn() = pg_last_wal_replay_lsn() THEN 0::int ELSE " \
                      "EXTRACT(EPOCH FROM now() - pg_last_xact_replay_timestamp())::int END AS time_lag;"
    replication_status_query = "SELECT status FROM get_replication_status();"

    def connect(self):
        """
        Connect to the database using psycopg2
        """
        self.connection = psycopg2.connect(
            database="dod_dbmon",
            user=self.user,
            password=self.password,
            host="/var/lib/pgsql/",
            port=self.port
        )

    def create(self, query):
        pass

    def get_replication_status(self):
        cursor = self.connection.cursor()
        cursor.execute(self.replication_status_query)
        results = cursor.fetchone()
        self.connection.commit()
        cursor.close()
        if isinstance(results, tuple):
            return results[0]
        else:
            return results

class StatusCodeException(Exception):
    """
    Raised when using InfluxDbodInstance, and result
    code was different than 200
    """


class InfluxDbodInstance(DbodInstance):
    """Represents a InfluxDB instance. Is a subclass of DbodInstance"""
    ping_insert_query = "collectd_ping,measurement=ping value={0}"
    ping_select_query = "q=select * from collectd_ping order by "\
        "time desc limit 1"
    ping_create_query = None
    ping_delete_query = None
    seconds_since_epoch = 0

    def connect(self):
        """
        Since we access InfluxDB with HTTP API there is no point
        in doing anything in the connect method
        """

    def select(self, statement):
        """
        Select writen point if Influx 1.x or ping command if Influx 2.x
        """
        if StrictVersion(self.db_version) >= StrictVersion('2.0'):
            # We call Influx 2.0 ping method
            # example: /usr/local/influxdb/influxdb-client-latest/bin/influx ping --host "https://dbod-xxx.cern.ch:8082"
            executable = "%s" % self.influx2_client
            param_host = "https://%s:%s" % (self.instance, self.port)
            out = subprocess.Popen([executable, 'ping', '--host', param_host],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = out.communicate()
            if stderr:
                raise StatusCodeException(
                    "Error pinging instance %s. Response code was: %s" %
                    (self.instance, stderr))
            else:
                return self.seconds_since_epoch
        else:
            url = "https://%s:%s/query?db=dod_dbmon" % (self.instance, self.port)
            response = requests.get(
                url, auth=(self.user, self.password), params=statement,
                verify=False, timeout=10)  # nosec
            if response.status_code != 200:
                raise StatusCodeException(
                    "Error executing GET statement on %s instance. Response "
                    "code was: %s" % (self.instance, response.status_code))
            return response.json()["results"][0]["series"][0]["values"][0][2]

    def insert(self, query):
        """
        Write a point if instance is running Influx 1.x
        """
        if StrictVersion(self.db_version) < StrictVersion('2.0'):
            # Influx uses HTTP API to write, here we POST to the influx write URL
            url = "https://%s:%s/write?db=dod_dbmon" % (self.instance, self.port)
            response = requests.post(
                url, auth=(self.user, self.password), data=query,
                verify=False, timeout=10)  # nosec
            if response.status_code != 204:
                raise StatusCodeException(
                    "Error executing POST statement on %s instance. Response "
                    "code was: %s" % (self.instance, response.status_code))
        else:
            self.seconds_since_epoch = int(query.split("value=")[1])

    def execute(self, statement):
        pass

    def create(self, query):
        pass

    def disconnect(self):
        """
        Also, here we won't do anything
        """


class MalformedDbodInstance(DbodInstance):
    """Represents a malformed instance, i.e. an instance that has some missing
    information in it's description (in entities.json). For example, such
    missing information makes connecting impossible."""
    # pylint: disable=super-init-not-called
    def __init__(self, api_dict, user=None, password=None):
        """
        Don't call parent class constructor, keep the dict, to see
            what was missing
        """
        self.api_dict = api_dict
        try:
            self.state = api_dict["state"]
        except KeyError:
            self.state = "UNKNOWN"

    def connect(self):
        pass

    def execute(self, statement):
        pass

    def disconnect(self):
        pass


instance_types = {    # pylint: disable=invalid-name
    "PG": PostgresDbodInstance,
    "MYSQL": MySQLDbodInstance,
    "InfluxDB": InfluxDbodInstance
}



import dbod_plugin
import pytest
import string
from hypothesis import given, strategies as st, settings, HealthCheck
from mock import call
from dbod_instances import MalformedDbodInstance
from helpers import get_mocked_instance


def test_dbod_plugin(mocker):
    """
    Test that DbodCollectdPluginPing is a subclass of DbodCollectdPlugin
    """
    mocker.patch('dbod_plugin.DbodCollectdPlugin.__init__')
    dbod_plugin.DbodCollectdPlugin.__init__.return_value = None
    from dbod_ping import DbodCollectdPluginPing
    i = DbodCollectdPluginPing()
    assert isinstance(i, dbod_plugin.DbodCollectdPlugin)


def test_read_calls_parent_read(mocker):
    """
    Parent class read must be called to refresh instances data
    """
    mocker.patch('dbod_plugin.DbodCollectdPlugin.__init__')
    mocker.patch('dbod_plugin.DbodCollectdPlugin.read')
    dbod_plugin.DbodCollectdPlugin.__init__.return_value = None
    from dbod_ping import DbodCollectdPluginPing

    plugin = DbodCollectdPluginPing()
    try:
        plugin.read()
    except: # noqa
        pass

    dbod_plugin.DbodCollectdPlugin.read.assert_called()


@pytest.fixture()
def plugin(mocker):
    """
    Returns an instance of DbodCollectdPlugin with
    all the real things mocked:
        - plugin.helper mocked - the constructor
            opens the config files to read passwords etc.
        - plugin.read mocked - it calls helper.get_instances which
            sends a request DBOD API to get list of entities
    """
    # We mock init of the parent class because it calls collectd methods
    # to register read
    mocker.resetall()
    init = mocker.patch('dbod_plugin.DbodCollectdPlugin.__init__')
    init.return_value = None
    # Read from parent class would call DBOD API to get instances
    # running locally, we provide our own instances in the tests
    mocker.patch('dbod_plugin.DbodCollectdPlugin.read')
    mocker.patch('dbod_plugin.DbodCollectdPlugin.dispatch')
    from dbod_ping import DbodCollectdPluginPing
    p = DbodCollectdPluginPing()
    # Mock instances attribute - this noramlly comes from the DBOD API

    mocker.patch.object(p, 'helper', create=True)
    return p


@given(
    instances=st.lists(
        # Create a list of 100 instance mocks
        st.builds(
            get_mocked_instance,
            db_name=st.integers(max_value=10),
            slave=st.booleans(),
            active=st.booleans(),
            malformed=st.booleans(),
            exception=st.sampled_from([None]),
            select_value=st.integers(
                min_value=1, max_value=10),
            notification=st.booleans()
            ),
        min_size=1,
        max_size=20
        ),
    # Simulate output of time.time() that we insert
    time_val=st.integers(min_value=1, max_value=10)
)
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_read_instance_methods(mocker, plugin, instances, time_val):
    """
    Test that depending on the instance attributes (ative, slave)
    correct methods are called
    Additionally, test that MalformedDbodInstance is not taken
    into account at all
    """
    #########################################
    # Mock various things
    #########################################
    # plugin.instances normally comes from DBOD API,
    # here we set it to mocks generated with hypothesis
    mocker.resetall()
    mocker.patch.object(plugin, 'instances', instances, create=True)
    # mock time call
    time = mocker.patch('dbod_ping.time')
    time.return_value = time_val
    ##################################################
    # The actual call that we're testing,
    # normally read is called by collectd!
    ##################################################
    plugin.read()
    # We iterate over mocked instances and check
    # that proper calls were made
    ##################################################
    for instance in instances:
        # Malformed instance shouldn't run anything
        if isinstance(instance, MalformedDbodInstance):
            instance.assert_not_called()
            continue
        # Not active instance or slave. We shouldn't connect
        if instance.slave or not instance.active:
            instance.assert_not_called()
            continue
        # We should create, insert and select to not slave instances
        calls = [call.connect(),
                 call.create(instance.ping_create_query),
                 call.execute(instance.ping_delete_query),
                 call.ping_insert_query.format(time_val),
                 call.insert(instance.ping_insert_query.format()),
                 call.select(instance.ping_select_query),
                 call.disconnect()]
        instance.mock_calls == calls


@given(
    instances=st.lists(
        # Create a list of 100 instance mocks
        st.builds(
            get_mocked_instance,
            db_name=st.integers(max_value=10),
            slave=st.booleans(),
            active=st.booleans(),
            malformed=st.booleans(),
            exception=st.sampled_from([None]),
            select_value=st.integers(
                min_value=1, max_value=10),
            notification=st.booleans()),
        min_size=1,
        max_size=20
        ),
    # Simulate output of time.time() that we insert
    time_val=st.integers(min_value=1, max_value=10)
)
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_read_dispatch(mocker, plugin, instances, time_val):
    """
    Test that depending on the instance attributes (ative, slave),
    raised exceptions and inserted/selected value
    correct value is dispatched to collectd
    """
    #########################################
    # Mock various things
    #########################################
    # plugin.instances normally comes from DBOD API,
    # here we set it to mocks generated with hypothesis
    mocker.resetall()
    mocker.patch.object(plugin, 'instances', instances, create=True)
    plugin.dispatch.reset_mock()
    # mock time call
    time = mocker.patch('dbod_ping.time')
    time.return_value = time_val
    ##################################################
    # The actual call that we're testing,
    # normally read is called by collectd!
    plugin.read()
    ##################################################
    # We collect all calls to dispatch here
    dispatch_calls = []
    # We iterate over mocked instances and check
    # that proper calls to dispatch were made
    ##################################################
    log_dispatch = "| dbod_ping | {} | {:10} | [{}] | {}"
    for instance in instances:
        # Malformed instance should be skipped in the ping plugin
        # there is another plugin for them
        if isinstance(instance, MalformedDbodInstance):
            continue
        # Not active instance we send -1
        if not instance.active:
            dispatch_calls.append(
                call([-1], instance.notification,
                    log_dispatch.format("INFO", instance.db_name,
                    "-1", "Instance not active."),
                    plugin_instance=instance.db_name))
            continue

        if not instance.slave:
            # If inserted value = selected value we dispatch 0,
            # otherwise we dispatch 1
            if time_val == instance.select_value:
                dispatch_calls.append(
                    call([0], instance.notification,
                        log_dispatch.format("INFO", instance.db_name,
                        "-1", "Notifications disabled"),
                        plugin_instance=instance.db_name))
            else:
                dispatch_calls.append(
                    call([1], instance.notification,
                        log_dispatch.format("ERROR", instance.db_name,
                        "-1", "Notifications disabled"),
                        plugin_instance=instance.db_name))

    assert plugin.dispatch.mock_calls == dispatch_calls

@given(
    instances=st.lists(
        # Create a list of 100 instance mocks
        st.builds(
           get_mocked_instance,
            db_name=st.integers(max_value=10),
            slave=st.booleans(),
            active=st.just(True),
            # No Malformed instances in this test
            malformed=st.just(False),
            # simulate exception in one of the methods
            exception=st.sampled_from(
                ["connect", "insert", None, "select", "create", "execute",
                 "disconnect"]),
            # Select always returns 100 as we're testing exceptions
            select_value=st.just(100)),
        min_size=1,
        max_size=20
        ),
    # Simulate output of time.time() that we insert
    time_val=st.just(100)
)
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_read_dispatch_exceptions(mocker, plugin, instances, time_val):
    """
    Test that if any of the instance methods
    (connect, insert, select, disconnect...)
    raises an exception, the plugin will dispatch a correct value
    """
    #########################################
    # Mock
    mocker.resetall()
    mocker.patch.object(plugin, 'instances', instances, create=True)
    plugin.dispatch.reset_mock()
    # mock time call
    time = mocker.patch('dbod_ping.time')
    time.return_value = time_val
    ##################################################
    # The actual call that we're testing,
    # normally read is called by collectd!
    plugin.read()
    ##################################################
    # We collect all calls to dispatch here
    dispatch_calls = []
    # We iterate over mocked instances and check
    # that proper calls to dispatch were made
    ##################################################
    log_dispatch = "| dbod_ping | {} | {:10} | [{}] | {}"

    for instance in instances:
        if not instance.active:
            dispatch_calls.append(
                call([-1], instance.notification,
                    log_dispatch.format("INFO", instance.db_name,
                    "-1", "Instance not active."),
                    plugin_instance=instance.db_name))
            continue

        if instance.slave:
            continue

        # if any method raises an exception we should dispatch 1
        # apart from the disconnect, which we consider not critical
        if instance.exception:
            if instance.exception != "disconnect":
                dispatch_calls.append(
                    call([1], instance.notification,
                        log_dispatch.format("ERROR", instance.db_name,
                        "-1", "Notifications disabled"),
                        plugin_instance=instance.db_name))
                continue

        dispatch_calls.append(
            call([0], instance.notification,
                log_dispatch.format("INFO", instance.db_name,
                "-1", "Notifications disabled"),
                plugin_instance=instance.db_name))

    assert plugin.dispatch.mock_calls == dispatch_calls


@given(
    instances=st.lists(
        # Create a list of 100 instance mocks
        st.builds(
            get_mocked_instance,
            db_name=st.integers(max_value=10),
            slave=st.booleans(),
            active=st.just(True),
            # No Malformed instances in this test
            malformed=st.booleans(),
            # simulate exception in one of the methods
            exception=st.sampled_from(
                ["connect", "insert", None, "select", "create", "execute",
                 "disconnect"]),
            # Select always returns 100 as we're testing exceptions
            select_value=st.integers(
                min_value=1,
                max_value=10
                )),
        min_size=1,
        max_size=20
        ),
    # Simulate output of time.time() that we insert
    time_val=st.integers(min_value=1, max_value=10)
)
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_read_updating_dbod_state(mocker, plugin, instances, time_val):
    """
    Test that instance state is correctly updated in the DBOD API
    """
    #########################################
    # Mock
    mocker.resetall()
    mocker.patch.object(plugin, 'instances', instances, create=True)
    plugin.helper.update_state.reset_mock()
    # mock time call
    time = mocker.patch('dbod_ping.time')
    time.return_value = time_val
    ##################################################
    # The actual call that we're testing,
    # normally read is called by collectd!
    plugin.read()
    ##################################################
    # We collect all calls to dispatch here
    update_state_calls = []
    # We iterate over mocked instances and check
    # that proper calls to update_state were made
    ##################################################
    from dbod_instances import DbodInstance
    for instance in instances:
        if isinstance(instance, MalformedDbodInstance):
            continue
        if instance.slave:
            continue

        # state in ping, only for master instances
        if not instance.exception or instance.exception == "disconnect":
            # If no exception was raise we should compare the inserted
            # and selected values. If they match, we update state to
            # RUNNING, otherwise to BUSY
            # also, f it failed to disconnect we should update to RUNNING
            if instance.select_value != time_val:
                update_state_calls.append(
                    call(instance, DbodInstance.BUSY))
            else:
                update_state_calls.append(
                    call(instance, DbodInstance.RUNNING))
            continue

        if instance.exception and instance.exception != "disconnect":
            # if any method raised an exception we should
            # update state to STOPPED
            # disconnect is not critical so we dont update the state
            update_state_calls.append(
                call(instance, DbodInstance.STOPPED))

    assert plugin.helper.update_state.mock_calls == update_state_calls


@given(
    instances=st.lists(
        # Create a list of 100 instance mocks
        st.builds(
            get_mocked_instance,
            db_name=st.integers(max_value=10),
            slave=st.booleans(),
            active=st.booleans(),
            # No Malformed instances in this test
            malformed=st.booleans(),
            # simulate exception in one of the methods
            exception=st.sampled_from(
                ["connect", "insert", None, "select", "create", "execute",
                 "disconnect"]),
            # Select always returns 100 as we're testing exceptions
            select_value=st.integers(
                min_value=1,
                max_value=10
                ),
            exception_message=st.text(
                alphabet=string.ascii_lowercase, max_size=30),
            notification=st.booleans(),
            ),
        min_size=1,
        max_size=20
        ),
    # Simulate output of time.time() that we insert
    time_val=st.integers(min_value=1, max_value=10)
)
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_read_collectd_error_messages(mocker, plugin, instances, time_val):
    """
    Test that before dispatching, we log a proper message so we can inspect
    later what happened in /var/log/collectd.log
    """
    #########################################
    # Mock
    mocker.resetall()
    import collectd
    mocker.patch.object(plugin, 'instances', instances, create=True)
    plugin.helper.update_state.reset_mock()
    collectd.reset_mock()
    # mock time call
    time = mocker.patch('dbod_ping.time')
    time.return_value = time_val
    ##################################################
    # The actual call that we're testing,
    # normally read is called by collectd!
    plugin.read()
    ##################################################
    # We collect all calls to dispatch here
    expected_calls_error = []
    expected_calls_info = []
    expected_calls_warning = []
    # We iterate over mocked instances and check
    # that proper calls to collectd.error were made
    ##################################################
    for instance in instances:
        if isinstance(instance, MalformedDbodInstance):
            continue
        if not instance.active:
            # If instance is not active, we should log that we dispatch -1
            expected_calls_info.append(
                call("| dbod_ping | INFO | %s | [%s] | %s"
                     % (instance.db_name, "-1", "Instance not active.")))
            continue
        if instance.slave:
            continue

        # Not slave instance
        # Any exception apart from disconnect method
        if instance.exception and \
                instance.exception != "disconnect":
            expected_calls_error.append(
                call("| dbod_ping | ERROR | %s | [%s] | %s"
                     % (instance.db_name, "1",
                        instance.exception_message)))
        elif time_val != instance.select_value:
            # Here we check that if inserted and selectd values
            # are different and we log it
            expected_calls_error.append(
                call("| dbod_ping | ERROR | %s | [%s] | %s"
                     % (instance.db_name, "1",
                        "Inserted value different than selected.")))
        if instance.exception == "disconnect":
            # Disconnect error should be logged
            expected_calls_warning.append(
                call("| dbod_ping | WARNING | %s | [] | %s"
                     % (instance.db_name,
                        "Not able to disconnect: %s"
                        % instance.exception_message)))

    assert collectd.error.mock_calls == expected_calls_error
    assert collectd.info.mock_calls == expected_calls_info
    assert collectd.warning.mock_calls == expected_calls_warning



from dbod_instances import DbodInstance, EntityMalformedException, \
    MySQLDbodInstance, PostgresDbodInstance, InfluxDbodInstance, \
    MalformedDbodInstance, StatusCodeException
from hypothesis import given, strategies as st, settings, HealthCheck
from pytest import raises
from mock import MagicMock, call
import dbod_instances
import pytest
import subprocess


class FakePopen(object):
    def __init__(self, args, stdout=None):
        super().__init__()
        self.args = args
        self.stdout = stdout

    def communicate(self):
        return "OK"


@pytest.fixture
def instance():
    """
    Dummy implementation of DbodInstance
    """
    class Instance(DbodInstance):
        def connect(self):
            pass

        def disconnect(self):
            pass

        def execute(self):
            pass

    return Instance


def test_dbod_instance_is_abstract():
    """
    DbodInstance is an abstract class, specific implementations
    include:
        - MySQLDbodInstance
        - PostgresDbodInstance
        - InfluxDbodInstance
        - MalformedDbodInstance - on missing data in dbod api
    """
    with raises(TypeError) as e:
        DbodInstance({})
    assert "Can't instantiate abstract class" in str(e.value)


def test_subclass_of_dbod_instance_must_implement(mocker):
    mocker.patch('dbod_instances.DbodInstance.__init__')
    DbodInstance.__init__.return_value = None

    class SubclassTest(DbodInstance):
        pass
    ##################################
    # connect()
    with raises(TypeError) as e:
        SubclassTest()
    assert "Can't instantiate abstract class SubclassTest with " \
        "abstract methods connect" in str(e.value)


def test_dbod_instance_init(test_instance1, mocker, instance):
    """
    Test it has all necessary fields
    """
    # To test what __init__ of DbodInstance does we create a subclass
    i = instance(test_instance1)
    # Name is concatenation of db_name
    assert i.instance == "dbod-test-db1.cern.ch"
    assert i.port == 6604
    assert i.db_name == "test_db1"
    assert i.db_type == "PG"
    assert i.socket == "/var/lib/pgsql/"
    assert i.category == "TEST"
    assert i.notification is True
    assert i.slave is False
    assert i.active is True
    assert i.state == "RUNNING"
    test_instance1["attributes"]["slave"] = True
    i = instance(test_instance1)
    assert i.slave is True
    test_instance1["attributes"]["notifications"] = True
    i = instance(test_instance1)
    assert i.notification is True
    # Assert connect and query methods are defined
    assert i.connect
    assert i.execute
    mocker.patch.object(i, 'execute')
    i.insert("insert sth")
    i.create("create sth")
    assert i.execute.mock_calls == [call("insert sth"), call("create sth")]

# Don't remove db_type, as it fails with the new API
@pytest.mark.parametrize(
    "key",
    ["db_name", "port", "socket", "class", "active", "state"])
def test_dbod_instance_init_missing_key(test_instance1, key, instance):
    """
    Test that it raises a proper exception on initialization
        EntityMalformedException
    Inside of this exception we should see what key was missing
    """
    del test_instance1[key]

    with raises(EntityMalformedException) as e:
        instance(test_instance1)
    assert str(e.value) == \
        "Attribute %s missing from the metadata! (DBOD API)" % key


@pytest.mark.parametrize(
    "key",
    ["notifications"])
def test_dbod_instance_init_missing_attribute(test_instance1, instance, key):
    """
    Test that it raises a proper exception on initialization
        EntityMalformedException
    Inside of this exception we should see what key was missing
    """
    del test_instance1["attributes"][key]

    with raises(EntityMalformedException) as e:
        instance(test_instance1)
    assert str(e.value) == \
        "Attribute %s missing from the metadata! (DBOD API)" % key


@pytest.mark.parametrize(
    "key,value,is_attr",
    [
        ("port", "1234gf", False),
        ("notifications", "notTrue", True),
        ("slave", "something", True)
    ])
def test_dbod_instance_init_incorrect_value_from_api(
        test_instance1, instance, key, value, is_attr):
    """
    Test that it raises a proper exception on initialization
        EntityMalformedException
    Inside of this exception we should see what key was missing
    """
    if is_attr:
        test_instance1["attributes"][key] = value
    else:
        test_instance1[key] = value

    with raises(EntityMalformedException) as e:
        instance(test_instance1)
    print(e.value)
    assert str(e.value).startswith(
        "Attribute %s has incorrect value! (DBOD API metadata)" % key)


def test_specific_implementation_exist(mocker):
    """
    Test that there are implementations of DbodInstance
    for each of the database, and they call the parent class __init__
    """
    mocker.patch('dbod_instances.DbodInstance.__init__')
    dbod_instances.DbodInstance.__init__.return_value = None
    some_dict = {}
    MySQLDbodInstance(some_dict)
    dbod_instances.DbodInstance.__init__.assert_called_with(some_dict)
    dbod_instances.DbodInstance.__init__.reset_mock()

    PostgresDbodInstance(some_dict)
    dbod_instances.DbodInstance.__init__.assert_called_with(some_dict)
    dbod_instances.DbodInstance.__init__.reset_mock()
    InfluxDbodInstance(some_dict)
    dbod_instances.DbodInstance.__init__.assert_called_with(some_dict)
    dbod_instances.DbodInstance.__init__.reset_mock()
    i = MalformedDbodInstance(some_dict)
    assert i.api_dict == some_dict
    dbod_instances.DbodInstance.__init__.assert_not_called()


def test_mapping_of_db_type_defined(mocker):
    """
    Test that there is a dict that maps db_type to class
    """
    from dbod_instances import instance_types
    assert instance_types["PG"] == PostgresDbodInstance
    assert instance_types["MYSQL"] == MySQLDbodInstance
    assert instance_types["InfluxDB"] == InfluxDbodInstance


def test_malformed_instance_keeps_state(mocker):
    """
    Test that MalformedInstance has state attribute
        - set to unknown in case of errors
    """
    api_dict = {}
    i = MalformedDbodInstance(api_dict)
    assert i.state == "UNKNOWN"
    api_dict = {"state": "AWAITING_APPROVAL"}
    i = MalformedDbodInstance(api_dict)
    assert i.state == "AWAITING_APPROVAL"


def test_dbod_instance_has_states():
    """
    Test that you can use DbodInstance.STOPPED
    """
    assert DbodInstance.STOPPED == "STOPPED"
    assert DbodInstance.RUNNING == "RUNNING"
    assert DbodInstance.BUSY == "BUSY"


######################################################
# POSTGRES
######################################################
def test_postgres_instance_has_plugin_queries():
    """
    PostgresDbodInstance should have:
        - ping_insert_query
        - ping_select_query
        - ping_create_query
        - ping_delete_query
        - slave_lag_query
    """
    assert PostgresDbodInstance.ping_insert_query ==\
        "INSERT INTO collectd_ping VALUES ({0});"
    assert PostgresDbodInstance.ping_select_query ==\
        "SELECT * FROM collectd_ping;"
    assert PostgresDbodInstance.ping_create_query is None
    assert PostgresDbodInstance.ping_delete_query ==\
        "DELETE FROM collectd_ping;"
    assert PostgresDbodInstance.slave_lag_query ==\
        "SELECT CASE WHEN pg_last_wal_receive_lsn() = pg_last_wal_replay_lsn() THEN 0::int ELSE " \
                "EXTRACT(EPOCH FROM now() - pg_last_xact_replay_timestamp())::int END AS time_lag;"

@pytest.mark.parametrize(
    "user,password,host,port",
    [
        ("user1", "pass1", "host-dbod1.cern.ch", "6060"),
        ("user2", "pass333", "host-dbod2.cern.ch", "60"),
        ("user3", "pass324", "host-dbod3.cern.ch", "16060")
    ])
def test_postgres_instance(mocker, user, password, host, port):
    """
    Test Postgres instance, it should have
        - connect()
        - insert(table, )
        - select()
        - create()
    """
    import psycopg2
    pginit = mocker.patch('dbod_instances.PostgresDbodInstance.__init__')
    pginit.return_value = None

    p = PostgresDbodInstance()
    mocker.patch.object(p, 'user', user, create=True)
    mocker.patch.object(p, 'password', password, create=True)
    mocker.patch.object(p, 'instance', host, create=True)
    mocker.patch.object(p, 'port', port, create=True)
    connect_result = MagicMock()
    psycopg2.connect.return_value = connect_result
    ############################################
    # Connect
    ############################################
    p.connect()

    psycopg2.connect.assert_called_with(
        database="dod_dbmon",
        user=p.user,
        password=p.password,
        host="/var/lib/pgsql/",
        port=p.port)
    assert p.connection == connect_result
    ############################################
    # Execute
    ############################################
    cursor = MagicMock()
    connect_result.cursor.return_value = cursor

    result = p.execute("INSERT SOMETHING")
    connect_result.cursor.assert_called()
    cursor.execute.assert_called_with("INSERT SOMETHING")
    connect_result.commit.assert_called()
    cursor.close.assert_called()
    assert result is None
    ############################################
    # Select
    ############################################
    cursor = MagicMock()
    connect_result.cursor.return_value = cursor
    fetchall = MagicMock()
    cursor.fetchall = fetchall
    fetchall.return_value = (["10"],)
    result = p.select("select SOMETHING")
    connect_result.cursor.assert_called()
    cursor.fetchall.assert_called()
    cursor.close.assert_called()
    assert result == "10"
    ############################################
    # Disconnect
    ############################################
    p.disconnect()
    connect_result.close.assert_called()


######################################################
# MySQL
######################################################

def test_mysql_instance_has_plugin_queries():
    """
    PostgresDbodInstance should have:
        - ping_insert_query
        - ping_select_query
        - ping_create_query
        - ping_delete_query
    """
    assert MySQLDbodInstance.ping_insert_query ==\
        "INSERT INTO collectd_ping VALUES ({0});"
    assert MySQLDbodInstance.ping_select_query ==\
        "SELECT * FROM collectd_ping;"
    assert MySQLDbodInstance.ping_create_query is None
    assert MySQLDbodInstance.ping_delete_query ==\
        "DELETE FROM collectd_ping;"
    assert MySQLDbodInstance.slave_lag_query ==\
        "SHOW SLAVE STATUS"


@pytest.mark.parametrize(
    "user,password,socket,port",
    [
        ("user1", "pass1", "/var/lib/host-dbod1.cern.ch", "6060"),
        ("user2", "pass333", "/var/lib/host-dbod2.cern.ch", "60"),
        ("user3", "pass324", "/var/lib/host-dbod3.cern.ch", "16060")
    ])
def test_mysql_instance(mocker, user, password, socket, port):
    """
    Test Postgres instance, it should have
        - connect()
        - insert(table, )
        - select()
        - create()
    """
    import pymysql
    pginit = mocker.patch('dbod_instances.MySQLDbodInstance.__init__')
    pginit.return_value = None

    p = MySQLDbodInstance()
    mocker.patch.object(p, 'user', user, create=True)
    mocker.patch.object(p, 'password', password, create=True)
    mocker.patch.object(p, 'socket', socket, create=True)
    mocker.patch.object(p, 'port', port, create=True)
    connection = MagicMock()
    pymysql.connect.return_value = connection
    ############################################
    # Connect
    ############################################
    p.connect()

    pymysql.connect.assert_called_with(
        unix_socket=p.socket,
        user=p.user,
        passwd=p.password,
        db="dod_dbmon",
        port=p.port,
        connect_timeout=10)
    assert p.connection == connection
    ############################################
    # Execute
    ############################################
    cursor = MagicMock()
    cursor.execute.return_value = "QUERY RESULT"
    connection.cursor.return_value = cursor

    result = p.execute("INSERT SOMETHING")
    connection.cursor.assert_called()

    cursor.execute.assert_called_with("INSERT SOMETHING")
    connection.commit.assert_called()
    cursor.close.assert_called()
    assert result is None
    ############################################
    # Select
    ############################################
    mocker.resetall()
    cursor = MagicMock()
    cursor.fetchall.return_value = (["QUERY RESULT"],)
    connection.cursor.return_value = cursor

    result = p.select("select SOMETHING")
    connection.cursor.assert_called()
    cursor.fetchall.assert_called()

    cursor.execute.assert_called_with("select SOMETHING")
    cursor.close.assert_called()
    assert result == "QUERY RESULT"

    ############################################
    # Disconnect
    ############################################
    p.disconnect()
    connection.close.assert_called()


def test_instance_accepts_user_and_pass(test_instance1, instance):
    i = instance(test_instance1, "user", "password")
    assert i.user == "user"
    assert i.password == "password"
    i = instance(test_instance1)
    with raises(AttributeError):
        i.user
    with raises(AttributeError):
        i.password


######################################################
# InfluxDB
######################################################

def test_influx_instance_has_plugin_queries():
    """
    InfluxDbodInstance should have:
        - ping_insert_query
        - ping_select_query
        - ping_create_query
    """
    assert InfluxDbodInstance.ping_insert_query ==\
        "collectd_ping,measurement=ping value={0}"
    assert InfluxDbodInstance.ping_select_query ==\
        "q=select * from collectd_ping order by time desc limit 1"
    assert InfluxDbodInstance.ping_create_query is None
    assert InfluxDbodInstance.ping_delete_query is None


@pytest.mark.parametrize(
    "user,password,instance,port,status_code,db_version,bindir",
    [
        ("user1", "pass1", "host-dbod1.cern.ch", "6060", 200, "1.8.2", "/usr/local/influxdb/influxdb-2.0.4"),
        ("user2", "pass333", "host-dbod2.cern.ch", "60", 200, "2.0.4", "/usr/local/influxdb/influxdb-2.0.4"),
        ("user3", "pass324", "host-dbod3.cern.ch", "16060", 203, "1.7.3", "/usr/local/influxdb/influxdb-2.0.4")
    ])
def test_influx_instance(mocker, user, password, instance, port, status_code, db_version, bindir):
    """
    Test Influx instance, it should have
        - connect()
        - insert(table, )
        - select()
        - create()
    It uses HTTP API to manipulat the data!
    """
    import requests
    init = mocker.patch('dbod_instances.InfluxDbodInstance.__init__')
    requests.get.return_value.status_code = status_code
    requests.get.return_value.json.return_value =\
        {"results": [{"series": [
            {"values": [["12/02/2019", "ping", 12345]]}]}]}
    init.return_value = None

    p = InfluxDbodInstance()
    mocker.patch.object(p, 'user', user, create=True)
    mocker.patch.object(p, 'password', password, create=True)
    mocker.patch.object(p, 'instance', instance, create=True)
    mocker.patch.object(p, 'port', port, create=True)
    mocker.patch.object(p, 'db_version', db_version, create=True)
    mocker.patch.object(p, 'bindir', bindir, create=True)
    ############################
    # Connect, shouldn't do anything
    ############################
    p.connect()
    requests.assert_not_called()
    p.execute("query")
    requests.assert_not_called()
    p.create("query")
    requests.assert_not_called()

    ######################################
    # select()
    ######################################
    if db_version[0] == "1":
        if status_code == 200:
            result = p.select("my query")
            assert result == 12345
        else:
            with raises(StatusCodeException) as exc:
                p.select("my query")
            assert "Error executing GET statement on %s instance."\
                " Response code was: %s" % (p.instance, status_code)\
                in str(exc.value)
        url = "https://%s:%s/query?db=dod_dbmon" % (instance, port)
        requests.get.assert_called_with(
            url, auth=(user, password), params="my query",
            verify=False, timeout=10)
    else:
        def fake_communicate(a):
            return "OK"
        subprocess.Popen = FakePopen
        subprocess.Popen.communicate = fake_communicate
        stdout = "12345"
        stderr = None
        if stderr:
            with raises(StatusCodeException) as exc:
                p.select("my query")
            assert "Error pinging instance %s. Response code was: %s"\
                % (p.instance, "err")
        else:
            assert stdout == "12345"

    ######################################
    # insert()
    ######################################
    # In case of insert 204 is the proper code
    status_code += 4
    requests.post.return_value.status_code = status_code
    if db_version[0] == "1":
        if status_code == 204:
            p.insert("my query")
        else:
            with raises(StatusCodeException) as exc:
                p.insert("my query")
            assert "Error executing POST statement on %s instance."\
                " Response code was: %s" % (p.instance, status_code)\
                in str(exc.value)
        url = "https://%s:%s/write?db=dod_dbmon" % (instance, port)
        requests.post.assert_called_with(
            url, auth=(user, password), data="my query",
            verify=False, timeout=10)
    else:
        p.insert("my query value=%s" % status_code)


def test_dbod_instance_has_get_replication_lag(mocker):
    """
    The basic implementation of get_replication_lag should just call select
    with slave_query
    """
    select = mocker.patch("dbod_instances.DbodInstance.select")
    init = mocker.patch("dbod_instances.DbodInstance.__init__")
    init.return_value = None
    select.return_value = 3

    class Test(DbodInstance):
        slave_lag_query = "select * from test"

        def connect(self):
            pass
    t = Test()
    assert t.get_replication_lag() == 3
    select.assert_called_with(t.slave_lag_query)


@pytest.mark.parametrize(
    "result,lag",
    [
        ("RESULT1", 2),
        (("Tuple result", 2), 10),
        (("Tuple", 3), 100)
    ])
def test_mysql_implementation_of_get_replication_lag(mocker, result, lag):
    init = mocker.patch("dbod_instances.DbodInstance.__init__")
    init.return_value = None
    izip = mocker.patch("dbod_instances.zip")
    izip.return_value = {"Seconds_Behind_Master": lag}
    i = MySQLDbodInstance()
    connection = mocker.patch.object(i, "connection", create=True)
    connection.cursor.return_value.fetchone.return_value = result
    connection.cursor.return_value.description = \
        (('Slave', 0, 1), ('Master', 0, 2))
    selected_lag = i.get_replication_lag()
    assert connection.mock_calls == [
        call.cursor(),
        call.cursor().execute(i.slave_lag_query),
        call.cursor().fetchone(),
        call.commit(),
        call.cursor().close()
    ]
    izip.assert_called_with(['Slave', 'Master'], result)
    assert lag == selected_lag


def merge_two_dicts(x, y):
    z = x.copy()   # start with x's keys and values
    z.update(y)    # modifies z with y's keys and values & returns None
    return z


@st.composite
def apidict(draw):
    """
    Generate dicitonary for a given entity, to simulate
    data as it comes from DBOD API or entities.json
    """
    api_dict = st.fixed_dictionaries(
        {
            "db_name": st.just("dbod"),
            "db_type": st.sampled_from(["PG", "MYSQL", "InfluxDB"]),
            "port": st.integers(1000, 3000),
            "socket": st.sampled_from(["/var/lib/mysql.socket",
                                       "/tmp/some.socket"]),
            "class": st.just("TEST"),
            "attributes": st.just({}),
            "active": st.booleans(),
            "state": st.sampled_from(["RUNNING", "BUSY", "STOPPED"]),
            "version": st.sampled_from(["1.7.3", "1.8.2", "2.0.4"]),
            "bindir": st.just("/usr/local/influxdb/influxdb-2.0.4")
            }
        )
    # Simulate missing attributes
    notifications = st.dictionaries(
        keys=st.just("notifications"),
        values=st.sampled_from(["true", "false"])
    )
    monit_notif_max_lag_seconds = st.dictionaries(
        keys=st.just("monit_notif_max_lag_seconds"),
        values=st.integers(100, 1000)
    )

    optional1 = draw(notifications)
    optional2 = draw(monit_notif_max_lag_seconds)

    api = draw(api_dict)
    api["attributes"] = merge_two_dicts(optional1, optional2)

    return api


@given(apidict())
@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
def test_max_lag_seconds(mocker, api_dict):
    """
    Test that DBOD Instance has max_lag_seconds attribute
    apidict() generates random test cases simulating dictionary
    as it comes from DBOD API or entities.json
    """
    mocker.resetall()
    class TestInstance(DbodInstance):
        def connect(self):
            pass

    # Fill missing pieces, that we're not interested in this test
    # api_dict will randomly make them empty or not defined
    api_dict["attributes"]["notifications"] = "true"
    api_dict["cluster_type"] = None
    api_dict["role"] = None
    t = TestInstance(api_dict)

    try:
        # If attribute defined in the dict, assert that object attribute is
        # set to proper value
        seconds = api_dict['attributes']['monit_notif_max_lag_seconds']
        assert t.max_lag_seconds == seconds
    except KeyError:
        # Assert default value was used
        assert t.max_lag_seconds == 300



"""Update instance status, get list of instances on a host

Raises:
    DbodConfigException: Missing configuration in  /etc/dbod/conf/core.conf
"""
import json
from socket import gethostname

import requests
from dbod_instances import EntityMalformedException, MalformedDbodInstance, \
    instance_types
import collectd  # pylint: disable=import-error

import platform
PYTHON_VERSION = platform.python_version()
from pkg_resources import parse_version as version

if version(PYTHON_VERSION) >= version("3.0"):
    from cerndb.config import Config
else:
    try:
        from ConfigParser import ConfigParser
    except ImportError:
        from configparser import ConfigParser
    import apacheconfig



class DbodConfigException(Exception):
    """Raised when the config is missing or malformed."""


class DbodCollectdHelper:
    """Class mostly used in performing operations with DBOD API."""

    # pylint: disable=unsubscriptable-object, too-many-instance-attributes
    # pylint: disable=import-outside-toplevel
    def __init__(self):
        if version(PYTHON_VERSION) >= version("3.0"):
            """Use cerndb-config"""

            try:
                self.config = Config( module_name="collectd-dbod")
            except Exception as e:
                raise DbodConfigException(e)
            ##################################################
            # Get interesting info from the config file /etc/collectd-dbod/config.yaml
            ##################################################
            try:
                self.cachefile = self.config['api']['cachefile']
                self.api_url = self.config['api']['host']
                self.instance_endpoint = \
                    self.config['api']['entity_endpoint']
                self.api_user = self.config['api']['user']
                self.api_password = self.config['api']['password']
                self.user = self.config['dbmon']['user']
                self.password = self.config['dbmon']['password']
                self.influx2_client = self.config['influx2']['client']
                self.cluster_member_endpoint = self.config['api']['cluster_member_endpoint']

            except KeyError as e:
                raise DbodConfigException(
                    "Error in DBOD config file (%s): missing key '%s'"
                    % ('config.yaml', e.args[0]))

            except TypeError as e:
                raise DbodConfigException(
                    "Error in DBOD config file (%s): missing nested key"
                    % ('config.yaml'))

        else:
            """Regethostnamead DBOD config file."""
            from apacheconfig import make_loader
            config_file = "/etc/dbod/conf/core.conf"
            ##################################################
            # Load DBOD config file to get api endpoints and
            # cachefile location
            ##################################################
            with make_loader() as loader:
                try:
                    self.config = loader.load(config_file)
                except apacheconfig.error.ConfigFileReadError as e:
                    raise DbodConfigException(e)
            ##################################################
            # Get interesting info from the config file
            ##################################################
            try:
                self.cachefile = self.config['api']['cachefile']
                self.api_url = self.config['api']['host']
                self.instance_endpoint = \
                    self.config['api']['entity_endpoint']
                self.api_user = self.config['api']['user']
                self.api_password = self.config['api']['password']
                self.user = self.config['dbmon']['user']
                self.password = self.config['dbmon']['password']
                self.influx2_client = self.config['influx2']['client']
                self.cluster_member_endpoint = \
                    self.config['api']['cluster_member_endpoint']

            except KeyError as e:
                raise DbodConfigException(
                    "Error in DBOD config file (%s): missing key '%s'"
                    % (config_file, e.args[0]))


    def get_instances(self):
        """
        Get running instances in the node using entities.json
        Returns:
            entities data as a dict (from /etc/dbod/conf/cache/entities.json)
        """
        ###################################
        # Get data from the cache: /etc/dbod/conf/cache/entities.json
        ###################################
        with open(self.cachefile) as f:
            data = json.load(f)

        ######################################
        # Create Instances from the response
        ######################################
        instances = []

        for instance_dict in data:
            # If creating DbodInstance raises exception, we either
            # pass it further or ignore it - depends on the metric!
            try:
                #####################################################
                # Create instances based on the db_type attribute
                #####################################################
                # Dynamically decide which instance
                # to create based on the db_type
                if instance_dict['state'] != 'MAINTENANCE':
                    try:
                        type = instance_dict['type']
                        instances.append(
                            instance_types[type](instance_dict,
                              user=self.user,
                              password=self.password,
                              influx2_client=self.influx2_client
                        ))

                    except KeyError:
                        # Missing db_type
                        instances.append(
                            MalformedDbodInstance(instance_dict,
                                                  user=self.user,
                                                  password=self.password))
            except EntityMalformedException:
                # Create MalformedDbodInstance
                instances.append(MalformedDbodInstance(
                    instance_dict, user=self.user, password=self.password))
        return instances

    def write_to_cache(self,db_name, key, value):
        with open(self.cachefile,'r') as f:
            data = json.load(f)
        for instance_dict in data:
            if instance_dict['name'] == db_name:
                instance_dict[key] = value
                break
        with open(self.cachefile,'w') as f:
            json.dump(data,f,indent=2,separators=(',', ': '))

    def update_state(self, instance, state):
        """
        instance - one of DbodInstance
        state - DbodInstance.STOPPED or DbodInstance.RUNNING for now

        update instance state in DBOD API
        """

        if instance.state == state:
            collectd.debug("Instance %s state [%s] has not changed from [%s]"
                           % (instance.db_name, instance.state, state))
        else:
            collectd.debug("Updating %s instance state to: %s"
                           % (instance.db_name, state))
            url = "%s/%s/%s" % (self.api_url,
                                self.instance_endpoint,
                                instance.id)
            # check the actual current value
            request = requests.get(url)
            # If status code != 200 there was an issue connecting to the DBOD api
            if request.status_code != 200:
                collectd.error("Failed getting current state of %s. Response code %s, url %s, self.api_url %s, self.instance_endpoint %s"
                               % (instance.db_name, request.status_code, url, self.api_url, self.instance_endpoint))
                # we then consider that the cache is good
                current_state = instance.state
            else:
               current_state = request.json()["response"][0]['state']
            # if the new instance state is one of these don't update
            if current_state in ["AWAITING_APPROVAL","MAINTENANCE", "TO_BE_DELETED"]:
                collectd.debug("Not updating state of %s. Instance state: %s"
                           % (instance.db_name, instance.state))
            elif current_state == state:
                collectd.debug("Instance %s state [%s] has not changed, but the cache needs to be updated"
                           % (instance.db_name, current_state))
                self.write_to_cache(instance.db_name, 'state', state)
            else:
                headers = {'auth': '{ "admin" : true , "groups" : [ ] , "owner" : "collectd" }' }
                response = requests.put(
                url, headers=headers,
                auth=(self.api_user, self.api_password),
                json={"state": state},
                verify=False)  # nosec
                if response.status_code != 204:
                    collectd.error("Failed updating state of %s. Response code %s, url %s, self.api_url %s, self.instance_endpoint %s, headers %s"
                               % (instance.db_name, response.status_code, url, self.api_url, self.instance_endpoint, headers))
                #also update the cache (DBOD-2744)
                self.write_to_cache(instance.db_name,'state', state)

    def update_instance_role(self, instance, role):
        """
        instance - one of DbodInstance
        role - PRIMARY or REPLICA for now
        update instance role in DBOD API
        """

        collectd.debug("Updating %s instance role to: %s"
                       % (instance, role))

        try:
            role_id = 1 if role == 'PRIMARY' else 2
            request_dict = {"instance_id": instance, "role_id": role_id}
            body = json.dumps(request_dict)
            headers = {'Content-Type': 'application/json',
                       'auth': '{ "admin" : true , "groups" : [ ] , "owner" : "collectd-agent" }'}
            url = '{base_path}/{cluster_member_endpoint}'.format(base_path=self.api_url,cluster_member_endpoint=self.cluster_member_endpoint)
            response = requests.put(url, data=body, headers=headers, auth=(self.api_user, self.api_password))

            if response.status_code == 200:
                collectd.debug(
                    "Instance  {node}  {role} attribute updated successfully".format(node=instance,
                                                                                          role=role))
            else:
                collectd.error("DBOD api returned code: {code}".format(code=response.status_code))
        except Exception as e:
            collectd.error("There was an error while updating {node} role. Error: {err}".format(node=instance, err=e))


    def update_instance_attribute(self, instance_name, attribute, value):
        """
        :param self:
        :param instance_name:
        :param attribute:
        :return:
        """
        collectd.debug("Updating {} '{}' attribute".format(instance_name, attribute))

        try:
            body = value
            headers = {'Content-Type': 'text/html; charset=utf-8',
                       'auth': '{ "admin" : true , "groups" : [ ] , "owner" : "collectd-agent" }'}
            url = '{base_path}/{entity_endpoint}/{node}/attribute/{attr}'.format(base_path=self.api_url, entity_endpoint=self.instance_endpoint,node=instance_name,
                                                                   attr=attribute)
            response = requests.put(url, data=body, headers=headers, auth=(self.api_user, self.api_password))

            if response.status_code == 200:
                collectd.debug(
                    "Instance  {node}  {attribute} attribute updated successfully".format(node=instance_name,
                                                                                          attribute=attribute))
            else:
                collectd.error("DBOD api returned code: {code}".format(code=response.status_code))
        except Exception as e:
            collectd.error("There was an error while updating {node} attribute. Error: {err}".format(node=instance_name, err=e))



"""
DBOD slave
"""

from dbod_plugin import DbodCollectdPlugin
from dbod_instances import DbodInstance, MalformedDbodInstance
import collectd  # pylint: disable=import-error


class DbodSlavePlugin(DbodCollectdPlugin):
    """
    DBOD slave plugin
    """

    log_dispatch = "| dbod_slave | {} | {:10} | [{}] | {}"

    def repl_lag_check(self,instance):

        collectd.debug("dbod_slave - checking replica {} replication lag".format(
                        instance.db_name))
        try:
            instance.connect()
            seconds = instance.get_replication_lag()
            # Compare the lag with the max_lag_seconds
            collectd.debug(
                "dbod_slave - comparing current lag ({}) with max_lag ({})".format(seconds, instance.max_lag_seconds))
            if type(seconds) is int:
                if seconds > instance.max_lag_seconds:
                    collectd.error(
                        self.log_dispatch.format(
                            "ERROR", instance.db_name, "1",
                            "Instance lags {} seconds behind master".format(seconds)))
                    self.dispatch([1], instance.notification,
                                  self.log_dispatch.format(
                                      "ERROR", instance.db_name,
                                      "-1", "Notifications disabled"),
                                  plugin_instance=instance.db_name)
                    self.helper.update_state(instance, DbodInstance.BUSY)
                else:
                    self.dispatch([0], instance.notification,
                                  self.log_dispatch.format(
                                      "INFO", instance.db_name,
                                      "-1", "Notifications disabled"),
                                  plugin_instance=instance.db_name)
                    self.helper.update_state(instance,
                                         DbodInstance.RUNNING)
            else:
                collectd.error(
                    self.log_dispatch.format(
                        "ERROR", instance.db_name, "1",
                        "Instance lag seconds behind primary is NULL"
                        ))
                self.dispatch([1], instance.notification,
                              self.log_dispatch.format(
                                  "ERROR", instance.db_name,
                                  "-1", "Notifications disabled"),
                              plugin_instance=instance.db_name)
                self.helper.update_state(instance, DbodInstance.BUSY)
        except Exception as exc:
            # Catch all exception and dispatch error
            collectd.error(
                self.log_dispatch.format(
                    "ERROR", instance.db_name, "1",
                    "Exception: {}" .format(exc)))
            self.dispatch([1], instance.notification,
                          self.log_dispatch.format(
                              "ERROR", instance.db_name,
                              "-1", "Notifications disabled"),
                          plugin_instance=instance.db_name)
            self.helper.update_state(instance, DbodInstance.STOPPED)
        finally:
            try:
                # Close connection to the database
                instance.disconnect()
            except Exception as exc:
                # We catch all exceptions - maybe the plugin
                # wasn't able to connect to the instance at all?
                collectd.warning(
                    self.log_dispatch.format(
                        "WARNING", instance.db_name, "",
                        "Not able to disconnect: {}".format(exc)))

    def read(self):
        """
        Call super - this will refresh list of instances from the Dbod API
        Then iterate over the list of instances, and check if they are slave
        """
        super(DbodSlavePlugin, self).read()
        for instance in self.instances:
            # MalformedDbodInstance means broken data in DBOD API
            # this is handled by another collectd plugin
            # Catch all exception and dispatch error
            if isinstance(instance, MalformedDbodInstance):
                continue
            if instance.cluster_type == 'InnoDB':
                continue
            if instance.slave:
                collectd.debug("-------")
                collectd.debug("dbod_slave - instance {} is a replica".format(instance.db_name))
                self.repl_lag_check(instance)



DbodSlavePlugin()



#!/usr/bin/env python
"""
setup.py configuration
"""

from setuptools import setup, find_packages

setup(
    name='collectd-dbod',
    version='1.9',
    description='Collectd plugins for DBoD',
    author='Damian Moskalik',
    author_email='damian.moskalik@cern.ch',
    url='https://gitlab.cern.ch/db/collectd/collectd-dbod',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[
        'dbod_entities_malformed',
        'dbod_helpers',
        'dbod_ping',
        'dbod_plugin',
        'dbod_instances',
        'dbod_slave',
        'dbod_innodb_cluster_ping',
        'dbod_replication'
    ],
    install_requires=[
        "cerndb-base",
        "apacheconfig",
        "requests",
        "psycopg2-binary",
        "psycopg2",
        "six",
        "pymysql",
        "setuptools"
    ],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
    ]
)



from pytest import raises
from mock import MagicMock
import dbod_helpers
import pytest
import collectd
from dbod_plugin import DbodCollectdPlugin # noqa


@pytest.fixture
def plugin(mocker):
    mocker.patch('dbod_helpers.DbodCollectdHelper.__init__')
    dbod_helpers.DbodCollectdHelper.__init__.return_value = None

    class Plugin(DbodCollectdPlugin):
        def read(self):
            super(Plugin, self).read()
            pass
    return Plugin


def test_class_is_abstract():
    """
    Test it's an abstract class with collectd_init and read methods
    """
    with raises(TypeError) as exc:
        DbodCollectdPlugin()
    assert "with abstract methods read" in str(exc.value)


def test_it_has_methods():
    """
    Test DbodCollectdPlugin has collectd_init() and read()
    methods
    """
    assert "collectd_init" in dir(DbodCollectdPlugin)
    assert "read" in dir(DbodCollectdPlugin)


def test_init_registers_methods_with_colelctd(plugin, mocker):
    """
    Constructor should register collectd_init() and read()
    """
    mocker.patch('dbod_helpers.DbodCollectdHelper.get_instances')
    dbod_helpers.DbodCollectdHelper.get_instances.return_value = []
    print(plugin)
    print(dbod_helpers.DbodCollectdHelper)
    print(dbod_helpers)
    print(DbodCollectdPlugin)
    p = plugin()
    collectd.register_read.assert_called_with(p.read)
    collectd.register_init.assert_called_with(p.collectd_init)

    dbod_helpers.DbodCollectdHelper.__init__.assert_called()
    assert p.collectd_type == "gauge"
    assert p.collectd_plugin == "test_dbod_collectd_plugin"
    assert isinstance(p.helper, dbod_helpers.DbodCollectdHelper)


def test_plugin_mock_read_calls_parent_read(mocker, plugin):
    """
    Test that read() will call get_instances
    """
    mocker.patch('dbod_plugin.DbodCollectdPlugin.read')
    p = plugin()
    p.read()
    DbodCollectdPlugin.read.assert_called()


def test_parent_read_refreshes_instances(mocker, plugin):
    """
    Constructor should register collectd_init() and read()
    """
    mocker.patch('dbod_helpers.DbodCollectdHelper.get_instances')
    dbod_helpers.DbodCollectdHelper.get_instances.return_value = []
    p = plugin()
    dbod_helpers.DbodCollectdHelper.get_instances.reset_mock()
    dbod_helpers.DbodCollectdHelper.get_instances.return_value \
        = ["test"]
    p.read()
    dbod_helpers.DbodCollectdHelper.get_instances.assert_called()
    assert p.instances == ["test"]


def test_collectd_init_calls_dbod_helper(mocker, plugin):
    """
    Test collectd_init method will get running instances
    """
    mocker.patch('dbod_helpers.DbodCollectdHelper.get_instances')
    dbod_helpers.DbodCollectdHelper.get_instances.return_value = []
    p = plugin()
    p.collectd_init()
    dbod_helpers.DbodCollectdHelper.__init__.assert_called()
    dbod_helpers.DbodCollectdHelper.get_instances.assert_called()
    assert p.instances == []
    assert isinstance(p.helper, dbod_helpers.DbodCollectdHelper)


def test_dispatch(mocker, plugin):
    """
    Test it has dispatch method, which accepts a list of values
        it should:
            - create collectd.Values
            - set values.type to gauge
            - set values.plugin to module.class
            - call values.dispatch
    """
    mock = MagicMock()
    mocker.patch("collectd.Values")
    collectd.Values.return_value = mock
    p = plugin()
    values = []
    p.dispatch(values)
    collectd.Values.assert_called()
    assert mock.type == "gauge"
    assert mock.plugin == "test_dbod_collectd_plugin"
    mock.dispatch.assert_called_with(values=values)


def test_dispatch_accepts_plugin_instance(mocker, plugin):
    """
    Test dispatch accepts plugin instance
    """
    mock_values = mocker.patch("collectd.Values")
    p = plugin()
    p.dispatch([1], plugin_instance="dbod1")
    assert mock_values.return_value.plugin_instance == "dbod1"
    mock_values.return_value.dispatch.assert_called_with(values=[1])
    p.dispatch([0], plugin_instance="dbod2")
    assert mock_values.return_value.plugin_instance == "dbod2"
    mock_values.return_value.dispatch.assert_called_with(values=[0])
    mock_values.return_value.plugin_instance = None
    p.dispatch([0])
    assert mock_values.return_value.plugin_instance is None
    mock_values.return_value.dispatch.assert_called_with(values=[0])
